from .Mesher import *
from .PatchMesher import *
from ..domain import Node, System
from ..elements.Element import Element


class TriPatchMesher(Mesher):
    """
    Mesher for plate and shell elements on a triangular domain,
    including in-plane and out-of plane loaded plates.
    """

    def __init__(self, model, *pts):
        super(TriPatchMesher, self).__init__(model, *pts)
        if len(pts)<3:
            raise TypeError("insufficient number of points given")

        X = list( pts[:3] )

        if len(pts) > 3 and pts[3]:
            X.append( pts[3] )
        else:
            Xm = 0.5*(np.array(X[0]) + np.array(X[1]))
            X.append( Xm )

        if len(pts) > 4 and pts[4]:
            X.append( pts[4] )
        else:
            Xm = 0.5*(np.array(X[1]) + np.array(X[2]))
            X.append( Xm )

        if len(pts) > 5 and pts[5]:
            X.append( pts[5] )
        else:
            Xm = 0.5*(np.array(X[2]) + np.array(X[0]))
            X.append( Xm )

        self.X = np.array(X)


    def quadMesh(self, Ne, element_type, material, **kwargs):
        """
        2D mesher using quadrilateral elements

        This is an abstract class - requires implementation in subclass

        :param int Ne: number of elements along each side
        :param element_type: compatible element type
        :param material: a material object
        :returns (nodes, elements):
        :var nodes: list of created :py:class:`Node` objects
        :var elements: list of created (subclass of) :py:class:`Element` objects
        """
        # if not element_type.isType(Element.QUAD):
        #     msg="Incompatible element type: {} in {}".format(element_type.__class__.__name__, sys._getframe().f_code.co_name)
        #     raise TypeError(msg)

        if Ne % 2:
            msg = "{} requires an even number elements per side. Ne={} was given.".format(self.__class__.__name__, Ne)
            raise TypeError(msg)

        # getting reference points

        # corners
        pt0 = self.map(0.0, 0.0)
        pt1 = self.map(1.0, 0.0)
        pt2 = self.map(0.0, 1.0)

        # mid points
        pt3 = self.map(0.5, 0.0)
        pt4 = self.map(0.5, 0.5)
        pt5 = self.map(0.0, 0.5)

        # center point
        pt6 = self.map(1./3., 1./3.)

        # quarter points on t=0
        pt7 = self.map(0.25, 0.0)
        pt8 = self.map(0.75, 0.0)

        # quarter points on u=0
        pt9 = self.map(0.75, 0.25)
        pt10 = self.map(0.25, 0.75)

        # quarter points on s=0
        pt11 = self.map(0.0, 0.75)
        pt12 = self.map(0.0, 0.25)

        # interior helper points
        pt13 = self.map(5./12., 2./12.)
        pt14 = self.map(5./12., 5./12.)
        pt15 = self.map(2./12., 5./12.)

        if self.model and isinstance(self.model, System):

            nodes    = []
            elements = []

            patch0 = PatchMesher(self.model, pt0, pt3, pt6, pt5, pt7, pt13, pt15, pt12)
            nds, elms = patch0.quadMesh(Ne//2, Ne//2, element_type, material)
            nodes += nds
            elements += elms

            patch0 = PatchMesher(self.model, pt1, pt4, pt6, pt3, pt9, pt14, pt13, pt8)
            nds, elms = patch0.quadMesh(Ne//2, Ne//2, element_type, material)
            nodes += nds
            elements += elms

            patch0 = PatchMesher(self.model, pt2, pt5, pt6, pt4, pt11, pt15, pt14, pt10)
            nds, elms = patch0.quadMesh(Ne//2, Ne//2, element_type, material)
            nodes += nds
            elements += elms

            self.model.addNode(*nodes)
            self.model.addElement(*elements)

        self.nodes    = nodes
        self.elements = elements

        return (nodes, elements)

    def triangleMesh(self, Ne, element_type, material, **kwargs):
        """
        2D mesher using triangular elements

        This is an abstract class - requires implementation in subclass

        :param int Ne: number of elements along each side
        :param element_type: compatible element type
        :param material: a material object
        :returns (nodes, elements):
        :var nodes: list of created :py:class:`Node` objects
        :var elements: list of created (subclass of) :py:class:`Element` objects
        """
        # if not element_type.isType(Element.TRIANGLE):
        #     msg="Incompatible element type: {} in {}".format(element_type.__class__.__name__, sys._getframe().f_code.co_name)
        #     raise TypeError(msg)

        # local coordinates
        s = np.linspace(0.,1.,Ne+1)
        t = np.linspace(0.,1.,Ne+1)

        nodes    = []
        elements = []

        for j in range(Ne+1):
            row = []
            for i in range(Ne+1-j):
                # mapping from local to global coordinates
                X, Y = self.map(s[i],t[j])
                nd = Node(X,Y)
                row.append(nd)
            nodes.append(row)

        for j in range(Ne):
            for i in range(Ne-j):
                node_list = [ nodes[j][i], nodes[j][i+1], nodes[j+1][i] ]
                elem = element_type(*node_list, deepcopy(material))
                elements.append(elem)
                if j>0:
                    node_list = [ nodes[j][i+1], nodes[j][i], nodes[j-1][i+1] ]
                    elem = element_type(*node_list, deepcopy(material))
                    elements.append(elem)

        nodes = [ nd for row in nodes for nd in row ]

        if self.model and isinstance(self.model, System):
            self.model.addNode(*nodes)
            self.model.addElement(*elements)

        self.nodes    = nodes
        self.elements = elements

        return (nodes, elements)

    def map(self,s,t):
        """
        maps the local coordinates (s,t) from a bi-unit-square
        to the actual position (x,y) in the global model space
        """
        u = 1. - s - t
        shp = [ u*(2*u-1), s*(2*s-1), t*(2*t-1), 4*u*s, 4*s*t, 4*t*u ]
        x = shp @ self.X + self.offset

        return (x[0], x[1])





import numpy as np

class PlotCurve():
    def __init__(self):
        self.x = []
        self.y = []
        self.z = []

    def __add__(self, Xt):
        if not isinstance(Xt, np.ndarray):
            raise TypeError

        if len(Xt)>=2:
            self.x.append(Xt[0])
            self.y.append(Xt[1])
        if len(Xt)>2:
            self.z.append(Xt[2])

        return self

    def asTuple(self):
        return (np.array(self.x), np.array(self.y), np.array(self.z))


class DrawElement():
    r"""
    class for drawing any element

    element types are:

    .. list-table::

       * - UNKNOWN
         -
       * - LINE
         - truss, simple frame
       * - CURVE
         - beam, nice frame
       * - TRIANGLE
         - plates, shells
       * - TETRAHEDRON
         - plates, shells
       * - QUAD
         - plates, shells
       * - BRICK
         - continuum

    """

    """
    defining element class parameters
    
    Add a new one for every general type of elements.
    """
    UNKNOWN     = 0x000000
    LINE        = 0x000001  # truss, simple frame
    CURVE       = 0x000002  # beam, nice frame
    TRIANGLE    = 0x000004  # plates, shells
    TETRAHEDRON = 0x000008  # plates, shells
    QUAD        = 0x000010  # plates, shells
    BRICK       = 0x000020  # continuum

    def __init__(self):
        self.element_type = self.UNKNOWN

    def isType(self, test_type):
        return (test_type & self.element_type)

    def draw(self, factor=0.0, modeshape=False, **kwargs):
        r"""
        Returns a series of coordinate vectors representing the _x_, _y_, and _z_ values
        of points for plotting the deformed shape the current element (:code:`self`).
        For the undeformed element, set :code:`factor=0.0`.

        :param factor:  magnification factor for deformation
        :returns:  a tuple (list) of coordinate vectors
        """
        kwargs['modeshape'] = modeshape

        if self.element_type == self.LINE:
            return self.drawLine(factor, **kwargs)
        elif self.element_type == self.CURVE :
            return self.drawCurve(factor, **kwargs)
        elif self.element_type == self.TRIANGLE :
            return self.drawTriangle(factor, **kwargs)
        elif self.element_type == self.TETRAHEDRON :
            return self.drawTetrahedron(factor, **kwargs)
        elif self.element_type == self.QUAD :
            return self.drawQuad(factor, **kwargs)
        elif self.element_type == self.BRICK :
            return self.drawBrick(factor, **kwargs)
        else:
            raise NotImplementedError


    def drawLine(self, factor, **kwargs):
        r"""
        implementation of a generic :code:`LINE` type
        """
        c = PlotCurve()
        if len(self.nodes) >= 2:
            for node in self.nodes[:2]:
                c += node.getDeformedPos(self, factor, **kwargs)
        return c.asTuple()

    def drawCurve(self, factor, **kwargs):
        r"""
        implementation of a generic :code:`CURVE` type
        """
        Xi = self.nodes[0].getPos()
        Xj = self.nodes[1].getPos()

        #
        Lvec = Xj - Xi
        L = np.linalg.norm(Lvec)
        n = Lvec / L

        if n.shape[0] == 2:
            # assuming 2D
            s = np.array([-n[1], n[0]])
        elif n.shape[0] == 3:
            # assuming 2D
            k = np.linalg.cross(n, np.array([0.1, np.pi, 0.0]))
            kk = np.linalg.norm(k)
            if kk <= 0.1:
                k = np.linalg.cross(n, np.array([-0.1, np.pi, 0.0]))
                kk = np.linalg.norm(k)
            k /= kk
            s = np.linalg.cross(kk, n)
        else:
            return tuple()

        # local displacements
        # 0 ... ux
        # 1 ... uy
        # 2 ... theta
        dispi = self.nodes[0].getDisp(dofs=('ux','uy','rz'), **kwargs)
        dispj = self.nodes[1].getDisp(dofs=('ux','uy','rz'), **kwargs)

        # ui     = dispi[0]    # the transformation was performed when we received dispi
        # vi     = dispi[1]    # the transformation was performed when we received dispi
        ui     = dispi[0:2] @ n
        vi     = dispi[0:2] @ s
        thetai = dispi[2] * L

        # uj     = dispj[0]    # the transformation was performed when we received dispi
        # vj     = dispj[1]    # the transformation was performed when we received dispj
        uj     = dispj[0:2] @ n
        vj     = dispj[0:2] @ s
        thetaj = dispj[2] * L

        qu = np.array([ui, uj]) * factor
        qv = np.array([vi, thetai, vj, thetaj]) * factor

        # shape functions
        xsi = np.linspace(0, 1, 11)
        phiU = [1. - xsi, xsi]
        phiV = [
            [1., 0.972, 0.896, 0.784, 0.648, 0.5, 0.352, 0.216, 0.104, 0.028, 0.],
            [0., 0.081, 0.128, 0.147, 0.144, 0.125, 0.096, 0.063, 0.032, 0.009, 0.],
            [0., 0.028, 0.104, 0.216, 0.352, 0.5, 0.648, 0.784, 0.896, 0.972, 1.],
            [0., -0.009, -0.032, -0.063, -0.096, -0.125, -0.144, -0.147, -0.128, -0.081, 0.]
        ]

        # deformed shape local
        xl = xsi * L
        xl += qu @ phiU
        yl = qv @ phiV

        # deformed shape global
        xvecs = Xi + np.outer(xl, n) + np.outer(yl, s)

        return (xvecs[:, 0], xvecs[:, 1])


    def drawTriangle(self, factor, **kwargs):
        r"""
        implementation of a generic :code:`TRIANGLE` type
        """
        c = PlotCurve()
        if len(self.nodes) >= 6:

            # 6-noded triangle

            c+= self.nodes[0].getDeformedPos(self, factor, **kwargs)
            c+= self.nodes[3].getDeformedPos(self, factor, **kwargs)
            c+= self.nodes[1].getDeformedPos(self, factor, **kwargs)
            c+= self.nodes[4].getDeformedPos(self, factor, **kwargs)
            c+= self.nodes[2].getDeformedPos(self, factor, **kwargs)
            c+= self.nodes[5].getDeformedPos(self, factor, **kwargs)
            c+= self.nodes[0].getDeformedPos(self, factor, **kwargs)

        elif len(self.nodes) >= 3:

            # 3-noded triangle

            for node in self.nodes[:3]:
                c += node.getDeformedPos(self, factor, **kwargs)
            c+= self.nodes[0].getDeformedPos(self, factor, **kwargs)

        return c.asTuple()

    def drawTetrahedron(self, factor, **kwargs):
        r"""
        implementation of a generic :code:`TETRAHEDRON` type
        """
        c = PlotCurve()
        if len(self.nodes) >= 4:
            for node in self.nodes:
                c += node.getDeformedPos(self, factor, **kwargs)
            c+= self.nodes[0].getDeformedPos(self, factor, **kwargs)
        return c.asTuple()

    def drawQuad(self, factor, **kwargs):
        r"""
        implementation of a generic :code:`QUAD` type
        """
        c = PlotCurve()
        match len(self.nodes):
            case 4:
                for node in self.nodes[:4]:
                    c += node.getDeformedPos(self, factor, **kwargs)
                c += self.nodes[0].getDeformedPos(self, factor, **kwargs)
            case 8:
                c += self.nodes[0].getDeformedPos(self, factor, **kwargs)
                c += self.nodes[4].getDeformedPos(self, factor, **kwargs)
                c += self.nodes[1].getDeformedPos(self, factor, **kwargs)
                c += self.nodes[5].getDeformedPos(self, factor, **kwargs)
                c += self.nodes[2].getDeformedPos(self, factor, **kwargs)
                c += self.nodes[6].getDeformedPos(self, factor, **kwargs)
                c += self.nodes[3].getDeformedPos(self, factor, **kwargs)
                c += self.nodes[7].getDeformedPos(self, factor, **kwargs)
                c += self.nodes[0].getDeformedPos(self, factor, **kwargs)
            case 9:
                c += self.nodes[0].getDeformedPos(self, factor, **kwargs)
                c += self.nodes[4].getDeformedPos(self, factor, **kwargs)
                c += self.nodes[1].getDeformedPos(self, factor, **kwargs)
                c += self.nodes[5].getDeformedPos(self, factor, **kwargs)
                c += self.nodes[2].getDeformedPos(self, factor, **kwargs)
                c += self.nodes[6].getDeformedPos(self, factor, **kwargs)
                c += self.nodes[3].getDeformedPos(self, factor, **kwargs)
                c += self.nodes[7].getDeformedPos(self, factor, **kwargs)
                c += self.nodes[0].getDeformedPos(self, factor, **kwargs)
                c += self.nodes[8].getDeformedPos(self, factor, **kwargs)
                c += self.nodes[0].getDeformedPos(self, factor, **kwargs)

        return c.asTuple()

    def drawBrick(self, factor, **kwargs):
        r"""
        implementation of a generic :code:`BRICK` type
        """
        c = PlotCurve()
        if len(self.nodes) >= 8:
            for node in self.nodes[:4]:
                c += node.getDeformedPos(self, factor, **kwargs)
            c+= self.nodes[0].getDeformedPos(self, factor, **kwargs)
        return c.asTuple()






import numpy as np
from copy import deepcopy

from ..Element import *
from ...domain.Node import *
from ...utilities import TriangleShapes, TriangleIntegration, GPdataType

class Triangle6(Element):
    r"""
    class: representing a 6-noded plane triangle
    """

    def __init__(self, node0, node1, node2, node3, node4, node5, material, label=None):
        super().__init__((node0, node1, node2, node3, node4, node5), material, label=label)

        ndim = node0.getPos().size

        if ndim == 3:
            dof_list = ('ux','uy','uz')
            ndof = 3
        elif ndim == 2:
            dof_list = ('ux','uy')
            ndof = 2
        else:
            raise TypeError("dimension of nodes must be 2 or 3")

        self.initialize(
            type=DrawElement.TRIANGLE,
            dofs=dof_list
                        )

        # select shape functions and integrator for the quadratic triangle
        self.interpolation = TriangleShapes()
        integrator    = TriangleIntegration(order=2)

        (self.xis, self.wis) = integrator.parameters()

        self.gpData = [ GPdataType() for i  in range(len(self.xis)) ]

        gpt = 0
        gp2nd_map = []

        for xi, wi, gpData in zip(self.xis,self.wis,self.gpData):

            gpData.material = deepcopy(material)

            shape   = self.interpolation.shape(  # requesting shape function array
                order=2,          # polynomial order per direction: quadratic
                s=xi[0], t=xi[1], # local coordinates for current position
                n=(0,0))          # no derivative with respect to (s,t)

            dshape1 = self.interpolation.shape(  # d shape function / d xi1
                order=2,          # polynomial order per direction: quadratic
                s=xi[0], t=xi[1], # local coordinates for current position
                n=(1,0))          # first derivative with respect to s

            dshape2 = self.interpolation.shape(  # d shape function / d xi2
                order=2,          # polynomial order per direction: quadratic
                s=xi[0], t=xi[1], # local coordinates for current position
                n=(0,1))          # first derivative with respect to t

            # covariant base vectors (reference system)
            X = np.zeros(ndim)
            base1 = np.zeros(ndim)
            base2 = np.zeros(ndim)
            for N, dN1, dN2, node in zip(shape, dshape1, dshape2, self.nodes):
                X     +=  N  * node.getPos()
                base1 += dN1 * node.getPos()
                base2 += dN2 * node.getPos()

            gpData.X = X
            gcov = np.vstack((base1, base2))
            gpData.base = gcov

            # metric (reference system)
            GIJ = gcov @ gcov.T
            gpData.state['GIJ'] = GIJ

            # dual base vectors (reference system)
            gpData.dual_base = np.linalg.inv(GIJ) @ gcov

            gpData.J = np.sqrt(np.linalg.det(GIJ))

            # populate gauss-point to nodes map
            raw_map = self.interpolation.shape(  # requesting shape function array
                order=1,  # polynomial order per direction: quadratic
                s=xi[0], t=xi[1],  # local coordinates for current position
                n=(0, 0))  # no derivative with respect to (s,t)
            map = raw_map * wi * gpData.J
            gp2nd_map.append(map)
            gpt += 1

        self.ngpts      = gpt                    # number of gauss points
        self._gp2nd_map = np.array(gp2nd_map).T  # gauss-point to nodes map array


    def __str__(self):
        s = super(Triangle6, self).__str__()
        for gp, gpdata in enumerate(self.gpData):
            s += "\n    strain {}: xx={xx:.3e} yy={yy:.3e} xy={xy:.3e} zz={zz:.3e}".format(gp, **gpdata.material.getStrain())
            s += "\n    stress {}: xx={xx:.3e} yy={yy:.3e} xy={xy:.3e} zz={zz:.3e}".format(gp, **gpdata.material.getStress())
        if np.array(self.distributed_load).any():
            s += "\n    element forces added to node:"
            for i, P in enumerate(self.Loads):
                Pi = np.array(P) * self.loadfactor
                s += "\n        {}: {}".format(self.nodes[i].getID(), Pi)
        return s

    def setSurfaceLoad(self, face, pn, ps=0):
        r"""
        .. list-table::
            :header-rows: 1

            * - face ID
              - nodes defining that face
            * - 0
              - :py:obj:`node 0` to :py:obj:`node 4` to :py:obj:`node 1`
            * - 1
              - :py:obj:`node 1` to :py:obj:`node 5` to :py:obj:`node 2`
            * - 2
              - :py:obj:`node 2` to :py:obj:`node 6` to :py:obj:`node 0`


        :param face: face ID for the laoded face
        :param pn: magnitude of distributed normal load per area. Tension on a surface is positive.
        :param ps: magnitude of distributed shear load per area. Positive direction is defined as shown in the above table.
        """
        if face >= 0 and face <= 2:
            self.faces[face].setLoad(pn, ps)

    def resetLoads(self):
        super(Triangle6, self).resetLoads()

    def updateState(self):

        # initializes internal force and tangent stiffness to zero arrays of the appropriate size.
        self.reset_matrices()

        for xi, wi, gpData in zip(self.xis,self.wis,self.gpData):

            # grab pre-computed quantities

            Gs = gpData.dual_base[0]
            Gt = gpData.dual_base[1]

            # shape functions

            dshape1 = self.interpolation.shape(  # d shape function / d xi1
                order=2,  # polynomial order per direction: quadratic
                s=xi[0], t=xi[1],  # local coordinates for current position
                n=(1, 0))  # first derivative with respect to s

            dshape2 = self.interpolation.shape(  # d shape function / d xi2
                order=2,  # polynomial order per direction: quadratic
                s=xi[0], t=xi[1],  # local coordinates for current position
                n=(0, 1))  # first derivative with respect to t

            # initialize covariant base vectors (current system) ...
            gs = np.zeros_like(Gs) #
            gt = np.zeros_like(Gt)

            # ... compute
            for dN1, dN2, node in zip(dshape1, dshape2, self.nodes):
                gs += dN1 * node.getDeformedPos(self)
                gt += dN2 * node.getDeformedPos(self)

            gu = -gs - gt

            # deformation gradient
            F = np.outer(gs, Gs) + np.outer(gt, Gt)

            # strain
            eps = 0.5 * ( F.T @ F - np.eye(np.size(gs)) )
            gpData.state['strain'] = eps

            # update the material state
            strain = {'xx':eps[0,0], 'yy':eps[1,1], 'xy':eps[0,1]+eps[1,0]}
            gpData.state['strain'] = strain
            gpData.material.setStrain(strain)

            # stress
            stress = gpData.material.getStress()

            S = np.array( [[stress['xx'],stress['xy']],[stress['xy'],stress['yy']]] )

            # 1st Piola-Kirchhoff stress
            P = F @ S

            # store stress for reporting
            gpData.state['stress'] = {'xx':P[0,0], 'xy':P[0,1], 'yx':P[1,0], 'yy':P[1,1]}

            S = np.array( [[stress['xx'],stress['xy']],[stress['xy'],stress['yy']]] )

            # initialize components of the B-matrix ...
            dshapeX = dshape1 * Gs[0] + dshape2 * Gt[0]
            dshapeY = dshape1 * Gs[1] + dshape2 * Gt[1]

            # initialize arrays
            gx = Gs[0] * gs + Gt[0] * gt
            gy = Gs[1] * gs + Gt[1] * gt

            # compute the kinematic matrices
            GI = np.vstack((dshapeX,dshapeY)).T

            BI = np.array([ [dNx*gx, dNy*gy, dNx*gy + dNy*gx] for (dNx,dNy) in zip(dshapeX,dshapeY) ])

            # internal force
            area = gpData.J * wi

            # stress * area
            stress_vec = np.array([stress['xx'], stress['yy'], stress['xy']]) * area

            # tangent stiffness
            Ct = gpData.material.getStiffness() * area

            for Krow, Fi, Gi, Bi in zip(self.Kt, self.Forces, GI, BI):
                Fi += Bi.T @ stress_vec
                KGi = Gi @ S * area
                for KIJ, Gj, Bj in  zip(Krow, GI, BI):
                    KGij = KGi @ Gj
                    KIJ += Bi.T @ Ct @ Bj + KGij * np.eye(np.size(gs))

        # .. applied element load (reference load)
        self.computeSurfaceLoads()

    def computeSurfaceLoads(self):
        r"""
        compute surface loads using faces

        This method should be called during :py:meth:`updateState()` by every
        element supporting surface loads

        """
        self.Loads = [ np.zeros_like(self.Forces[I]) for I in range(len(self.nodes)) ]

        for I, face in enumerate(self.faces):
            loads = face.computeNodalForces()

            # indexing: face-ID matches the start node
            M = I+3  # the middle node
            J = I+1  # the end node
            if J>2:
                J -= 3

            # add to element load vectors
            self.Loads[I] += loads[0]
            self.Loads[M] += loads[1]
            self.Loads[J] += loads[2]


    def getStress(self):
        stress = [ data.state['stress'] for data in self.gpData ]
        return stress

    def mapGaussPoints(self, var):
        r"""
        Initiate mapping of Gauss-point values to nodes.
        This method is an internal method and should not be called by the user.
        Calling that method explicitly will cause faulty nodal values.

        :param var: variable code for a variable to be mapped from Gauss-points to nodes
        """
        stresses = ('sxx','syy','szz','sxy','syz','szx')
        membrane = ('nxx','nyy','nxy')
        strains  = ('epsxx','epsyy','epszz','epsxy','epsyz','epszx')

        values = np.zeros( self.ngpts )

        if var.lower() in stresses:
            key = var[1:3].lower()
            values = []
            for gpdata in self.gpData:   # gauss-point loop
                if key in gpdata.state['stress']:
                    thickness = gpdata.material.getThickness()
                    values.append(gpdata.state['stress'][key]/thickness)
                else:
                    values.append(0.0)

        if var.lower() in membrane:
            key = var[1:3].lower()
            values = []
            for gpdata in self.gpData:   # gauss-point loop
                if key in gpdata.state['stress']:
                    values.append(gpdata.state['stress'][key])
                else:
                    values.append(0.0)

        if var.lower() in strains:
            key = var[1:3].lower()
            values = []
            for gpdata in self.gpData:   # gauss-point loop
                if key in gpdata.state['strain']:
                    values.append(gpdata.state['strain'][key])
                else:
                    values.append(0.0)

        values = np.array(values)

        for node, map in zip(self.nodes, self._gp2nd_map):
            wndi   = np.sum(map)
            val_wi = map @ values
            node._addToMap(wndi, val_wi)



"""
Example plate02
"""
from ...examples.Example import *

from ...domain.System import *
from ...solver.NewtonRaphsonSolver import *
from ...elements.LinearTriangle import *
from ...materials.PlaneStress import *


class ExamplePlate02(Example):

    def docString(self):
        s = """
    ## A square patch made of two triangular plate elements

    Basic implementation test with applied loads.
    Testing the tangent stiffness computation. 

    free   free
     ^     ^
     |     |
     3-----2 -> free
     |\  b | >
     | \   | >
     |  \  | > (w = 1.0)
     |   \ | >
     | a  \| >
     0-----1 -> free
    
    width:  10.
    height: 10.
    
    Material parameters: St. Venant-Kirchhoff, plane stress
        E  = 10.0
        nu =  0.30
        t  =  1.0
        
    Element loads:
        node 0: [ 0.0, 0.0]
        node 1: [ 5.0, 0.0]
        node 2: [ 5.0, 0.0]
        node 3: [ 0.0, 0.0]
    
    2nd Piola-Kirchhoff stress:
        S_XX =  w                  =  1.000
        S_YY = S_XY = S_YX = S_ZZ  =  0.000
        
    Green Lagrange strain:
        eps_XX = (1/E) ((1.000) - (0.30)(0.000)) =  0.100
        eps_YY = (1/E) ((0.000) - (0.30)(1.000)) = -0.030
        eps_XY = eps_YX                          =  0.000
        eps_ZZ = -nu * (eps_XX + eps_YY)         = -0.021
        
    Stretches:
        lam_X = sqrt(1 + 2 eps_XX) = 1.095
        lam_Y = sqrt(1 + 2 eps_YY) = 0.9695
        
    Displacements:
        ux = (lam_X - 1) * x, uy = (lam_Y - 1) * y
        node 0: [ 0.000,  0.000 ]
        node 1: [ 0.954,  0.000 ]
        node 2: [ 0.954, -0.305 ]
        node 3: [ 0.000, -0.305 ]
        
    Author: Peter Mackenzie-Helnwein 
    """
        return s

    def problem(self):

        params = dict(
            E  = 10.,   # Young's modulus
            nu = 0.3,   # Poisson's ratio
            t  = 1.0,   # thickness of the plate
            fy = 1.e30  # yield stress
        )

        a = 10.     # length of the plate in the x-direction
        b = 10.     # length of the plate in the y-direction

        model = System()
        model.setSolver(NewtonRaphsonSolver())

        nd0 = Node( 0.0, 0.0)
        nd1 = Node(   a, 0.0)
        nd2 = Node(   a,   b)
        nd3 = Node( 0.0,   b)

        nd0.fixDOF('ux', 'uy')
        nd1.fixDOF('uy')
        nd3.fixDOF('ux')

        model.addNode(nd0, nd1, nd2, nd3)

        elemA = LinearTriangle(nd0, nd1, nd3, PlaneStress(params))
        elemB = LinearTriangle(nd2, nd3, nd1, PlaneStress(params))

        model.addElement(elemA, elemB)

        elemB.setSurfaceLoad(face=2, w=1.0)

        model.setLoadFactor(0.0)
        model.solve()
        #model.report()  # activate this line for lots of debug info
        model.plot(factor=1.0,
                   title="Undeformed system",
                   filename="plate02_undeformed.png")

        model.setLoadFactor(1.0)
        model.solve()
        model.plot(factor=1.0, filename="plate02_deformed.png")

        model.report()




import numpy as np


class ShapeFunctions():
    """
    Abstract parent class for shape functions or their
    n-th (mixed) derivative for various interpolation functions.

    """

    UNKNOW    = 0x0000
    BEAMS     = 0x1000
    QUADS     = 0x2000
    TRIANGLE  = 0x4000
    BRICK     = 0x8000

    CONST     = 0x0001
    LINEAR    = 0x0002
    QUADRATIC = 0x0004
    CUBIC     = 0x0008
    OTHER     = 0x0010

    def __init__(self, type):
        self.type = type

    def shape(self, order, *argvs, **kwargs):
        """
        interface function to gets shape functions or their
        n-th derivative for various interpolation functions.

        .. note::

           This is a virtual function that needs to be implemented
           by any subclass

        """
        msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        raise NotImplementedError(msg)


def selfTestShapes():
    for order in range(5):
        print(f"order: {order}")
        for derivative in range(order+2):
            res = np.array( [ LineShapes().shape(order,xi,1.0,derivative) for xi in np.linspace(0,1,5) ] )
            print(res.T)
        print()

if __name__ == "__main__":
    selfTestShapes()



from .Material import Material
import numpy as np

class DiffusionGeneral(Material):
    """
    Generic diffusion model for 2D and 3D diffusion problems.

    Use subclasses :py:class:`Thermal`, :py:class:`Seapage`, etc., for actual analyses.
    """
    
    def __init__(self, params={'diffusivity':1., 'capacity':1., 'density':1., 'thickness':1.}):
        super(DiffusionGeneral, self).__init__(params)

        self._type = self.DIFFUSION

        if 'diffusivity' not in self.parameters:
            self.parameters['diffusivity'] = 1.0
        if 'capacity' not in self.parameters:
            self.parameters['capacity'] = 1.0
        if 'density' not in self.parameters:
            self.parameters['density'] = 1.0
        if 'thickness' not in self.parameters:
            self.parameters['thickness'] = 1.0

        self.gradPhi = None
        self.flux    = None

    def setGrad(self, gradPhi):
        """
        :param gradPhi: gradient :math:`\\nabla \\phi` of the scalar potential :math:`\\phi`
        :type gradPhi: numpy.ndarray
        """
        if isinstance(gradPhi, np.ndarray):
            self.gradPhi = gradPhi
            self.updateState()
        else:
            raise TypeError

    def getGrad(self):
        """
        :returns gradPhi: gradient :math:`\\nabla \\phi` of the scalar potential :math:`\\phi`
        :type gradPhi: numpy.ndarray
        """
        if isinstance(self.gradPhi,np.ndarray):
            return self.gradPhi
        else:
            return np.zeros(3)

    def getFlux(self):
        """
        :return: flux (numpy.ndarray)
        """
        if isinstance(self.flux,np.ndarray):
            return self.flux
        else:
            return np.zeros(3)

    def getThickness(self):
        """
        Returns thickness of a 2D plate, or 1.0 if none given or 3D problem
        """
        return self.parameters['thickness']

    def getCapacity(self):
        """
        returns :math:`C := \\varrho c`
        """
        return self.parameters['capacity']*self.parameters['density']

    def getDiffusivity(self):
        """
        Returns diffusivity of a 2D plate, or 1.0 if none given
        """
        return self.parameters['diffusivity']

    def updateState(self):
        """
        Update the material state after :math:`\\nabla \\phi`
        """
        if isinstance(self.gradPhi, np.ndarray):
            self.flux = -self.parameters['diffusivity'] * self.gradPhi




import sys
import numpy as np
import scipy as sc

import matplotlib.pyplot as plt

class Solver():
    r"""
    Abstract class for any solver implementation.

    This class describes the functions needed by any solver
    """

    def __init__(self):
        r"""
        Initialize a solver instance with empty elements and nodes lists.

        Provides a link to the model
        """
        self.loadfactor    = 1.0    # holds the current load factor

        self.loadfactor_n  = 0.0    # load factor for previously converged state
        self.loadfactor_nn = 0.0    # load factor for two steps back converged state

        self.model_ptr   = None     # establishes link to parent model
        self.elements    = []       # list of element pointers
        self.nodes       = []       # list of node pointers
        self.constraints = []       # list of constraint pointers
        self.sdof = 0               # number of DOFs in the current system

        # numeric iteration tolerance
        self.TOL = 1.0e-6

        # displacement control
        self.hasConstraint = False
        self.targetU       = 0.0

        # arc-length control
        self.useArcLength  = False
        self.lastConverged = {}

        # shall result be recorded?
        self.record = False

    def connect(self, model, nodes, elems, constraints):
        self.model_ptr   = model
        self.nodes       = nodes
        self.elements    = elems
        self.constraints = constraints

    def fetchState(self):
        r"""
        Fetch the current :code:`state` of the solver.

        .. list-table:: **state** is defined as a dictionary with the following contents:

            * - **model**
              - pointer to the linked model (required)
            * - **nodes**
              - list of node pointers (required)
            * - **elements**
              - list of element pointers (required)
            * - **P0**
              - system vector of initial forces
            * - **Pref**
              - system vector of reference forces
            * - **u1**
              - system vector of current (converged) displacements
            * - **un**
              - system vector of previous (converged) displacements
            * - **lam1**
              - load level of current (converged) displacements
            * - **lamn**
              - load level of previous (converged) displacements


        :return: state of the solver
        """
        state = {}
        state['model']    = self.model_ptr
        state['nodes']    = self.nodes
        state['elements'] = self.elements
        state['lam1']     = self.loadfactor

        return state

    def pushState(self, state):
        r"""
        Pushes :code:`state` to the solver.
        The solver will use that data to update it's internal state.

        .. list-table:: **state** is defined as a dictionary with the following contents:

            * - **model**
              - pointer to the linked model (required)
            * - **nodes**
              - list of node pointers (required)
            * - **elements**
              - list of element pointers (required)
            * - **P0**
              - system vector of initial forces
            * - **Pref**
              - system vector of reference forces
            * - **u1**
              - system vector of current (converged) displacements
            * - **un**
              - system vector of previous (converged) displacements
            * - **lam1**
              - load level of current (converged) displacements
            * - **lamn**
              - load level of previous (converged) displacements

        :param state: state of the solver
        """
        if 'model' in state:
            self.model_ptr = state['model']
        else:
            raise TypeError("'model' missing from state")

        if 'nodes' in state:
            self.nodes = state['nodes']
        else:
            raise TypeError("'nodes' missing from state")

        if 'elements' in state:
            self.elements = state['elements']
        else:
            raise TypeError("'elements' missing from state")

        if 'lam1' in state:
            self.loadfactor = state['lam1']
        else:
            raise TypeError("'lam1' missing from state")

    def setLoadFactor(self, lam):
        r"""
        Set the target load factor to **lam**

        This enables load control and disables alternative control methods
        such as *displacement control* or *arc-length* control.

        The entered load pattern is considered a reference load,
        to be multiplied by a scalar load factor, :math:`\lambda`.

        If no load factor is set explicitly, a factor of 1.0 is assumed, i.e., the
        entire entered load is applied in full.
        """
        self.loadfactor    = lam
        self.hasConstraint = False

    def setDisplacementControl(self, node, dof, target):
        r"""
        activate displacement control for the next load step
        """
        if dof in node.dofs:
            self.hasConstraint = True
            self.control_node  = node
            self.control_dof   = dof
            self.targetU       = target
        else:
            self.hasConstraint = False

            msg = f"{node} has no dof:{dof} to establish displacement control"
            raise TypeError(msg)

    def assemble(self, force_only=False):
        r"""
        A general assembler for mixed element types.

        This method will build the out-of-balance force vector (residuum :math:`{\bf R}`)
        and the tangent stiffness matrix (:math:`{\bf K}_t`) used by most solvers.

        Specialized solvers may overload this method.

        .. note::

            The solver will apply the global load factor to the reference load returned by nodes
            and elements.  While nodes and elements are aware of that load factor, they shall
            apply it **only if asked explicity** by passing :code:`apply_load_factor=True` to the
            respective access functions.

        :param force_only: set to **True** if only the residual force needs to be assembled
        """

        # compute size parameters
        ndof = 0
        for node in self.nodes:
            if node.isLead():
                node.setStart(ndof)
                ndof += node.ndofs

        for constraint in self.constraints:
            constraint.setStart(ndof)
            ndof += constraint.countConditions()

        self.sdof = ndof  # number of system d.o.f.s

        Psys = np.zeros(ndof)           # reference load vector (without load factor)
        Fsys = np.zeros(ndof)           # system internal force vector
        Ksys = np.zeros((ndof, ndof))   # system tangent stiffness matrix

        # displacement control or arc-length control
        if self.hasConstraint:
            if self.useArcLength:
                # arc-length control
                dload = self.loadfactor - self.loadfactor_n
                self.g = self.arclength2 - self.alpha * dload*dload * self.P@self.P
                for node in self.nodes:
                    self.g -= node.getNormDeltaU2()
            else:
                self.g = self.control_node.getDisp(self.control_dof) - self.targetU

        # assemble loads
        for node in self.nodes:
            if node.isLead() and node.hasLoad():
                # idx = node.start + np.arange(node.ndofs)
                idx = node.getIdx4DOFs()
                Psys[idx] += node.getLoad()

        # Element Loop: assemble element forces and stiffness
        for element in self.elements:

            Fe = element.getForce()     # Element State Update occurs here
            Pe = element.getLoad()      # Element State Update occurs here

            if not force_only:
                Ke = element.getStiffness() # fetch element stiffness matrix as array of nodal matrices

            for (i,ndI) in enumerate(element.nodes):
                # fetch dof mapping for node I
                idxK = ndI.getIdx4Element(element)

                # system reference load vector
                if isinstance(Pe[i], np.ndarray):
                    # assemble the load vector
                    Psys[idxK] += Pe[i]

                # system residual force vector
                Fsys[idxK] += Fe[i]

                # system tangent stiffness matrix
                if not force_only:
                    for (j,ndJ) in enumerate(element.nodes):
                        # fetch dof mapping for node J
                        idxM = ndJ.getIdx4Element(element)
                        # add to system matrix
                        Ksys[idxK[:, np.newaxis], idxM] += Ke[i][j]

        # system residual force vector
        self.P = Psys
        self.R = self.loadfactor * Psys - Fsys

        # apply boundary conditions
        if not force_only:
            for node in self.nodes:
                for dof in node.dofs:
                    if node.isFixed(dof):
                        #idx = node.lead.start + node.dofs[dof]
                        idx = node.getIdx4DOFs(dofs=[dof])[0]

                        ubar = node.getDisp(dof, local=True)

                        self.R[idx]    = ubar * 1.0e3
                        Ksys[:, idx]   = np.zeros(ndof)   # the range might need adjustment for constraints
                        Ksys[idx, :]   = np.zeros(ndof)   # the range might need adjustment for constraints
                        Ksys[idx, idx] = 1.0e3

            self.Kt = Ksys

    def solve(self, **kwargs):
        """
        """
        msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        raise NotImplementedError(msg)

    def initialize(self):
        """
        """
        msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        raise NotImplementedError(msg)

    def reset(self):
        """
        """
        msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        raise NotImplementedError(msg)

    def on_converged(self):
        r"""
        This function needs to be called once a converged state was achieved by the solver.

        It tells all components to update its state to "converged"
        """
        for node in self.nodes:
            node.setLoadFactor(self.loadfactor)
            node.on_converged()

        for elem in self.elements:
            elem.setLoadFactor(self.loadfactor)
            elem.on_converged()

        for const in self.constraints:
            const.setLoadFactor(self.loadfactor)
            const.on_converged()

        self.loadfactor_nn = self.loadfactor_n
        self.loadfactor_n  = self.loadfactor

    def revert(self):
        r"""
        This function needs to be called if the iterative procedure fails to converge.

        It will revert the entire system to the last converged state.
        """
        for node in self.nodes:
            node.revert()

        for elem in self.elements:
            elem.revert()

        for const in self.constraints:
            const.revert()

    def checkStability(self, verbose=True, **kwargs):
        r"""
        Computes the stability index as

        * :math:`\mathop{det}([{\bf K}_t])` for systems with less than 25 d.o.f.s
        * :math:`\min\lambda_i` where :math:`\lambda_i` are the eigenvalues of :math:`{\bf K}_t`

        :param verbose: set to **True** for log info
        :param num_eigen: if set to a value greater than 0, show the **num_eigen** eigenvalues
                        closest to the current load level.
        :returns: stability index
        """
        if 'num_eigen' in kwargs:
            num_eigen = kwargs['num_eigen']
        else:
            num_eigen = 0

        if self.sdof < 10 and not num_eigen:
            detKt = np.linalg.det(self.Kt)
            msg = f"\n ** Stability check: det(Kt) = {detKt}\n"
        else:
            evals = sc.linalg.eigvalsh(self.Kt)
            abs_evals = np.abs(evals)
            idx = np.argmin(abs_evals)
            detKt = evals[idx]  # we need to get the sign back
            if num_eigen > 0:
                detKt = []
                for cnt in range(num_eigen):
                    detKt.append(evals[idx])      # we need to get the sign back
                    abs_evals[idx] = 1.0e20
                    idx = np.argmin(abs_evals)
                detKt.sort()
                msg = f"\n ** Stability check: (smallest {num_eigen} eigenvalues of Kt)\n"
                for k, val in enumerate(detKt):
                    msg += f"\t\t\tmode {k}:{val:12.2f}\n"
            else:
                msg = f"\n ** Stability check: (smallest eigenvalue of Kt) = {detKt}\n"

        if verbose:
            print(msg)

        return detKt

    def getBucklingMode(self, mode=0, **kwargs):
        r"""
        Perform an eigen-analysis on :math:`{\bf K}_t` for the requested **mode**.
        Default is the mode with the smallest absolute eigenvalue (:math:`\min\{\lambda_i\}`)

        The mode shape will be pushed to the nodes.

        .. note::

            Use **System.plotBucklingMode()** for plotting of the eigenmode (see :doc:`../Domain/System_class`).
            This plotting method will be replaced by :code:`System.plot(...)` in a future release.

        :return: the eigenvalue, :math:`\lambda_{\mathtt{mode}}`
        """
        if not isinstance(mode,int) or mode < 0 or mode >= self.Kt.shape[0]:
            raise TypeError(f"mode out of range: must be an int between 0 and the number of d.o.f.s")

        w, v = sc.linalg.eigh(self.Kt, subset_by_index=[mode, mode])
        lam = w[0]
        U = v[:,0]

        # update nodal displacements
        for node in self.nodes:
            idxK = node.getIdx4DOFs()
            node.setDisp(U[idxK], modeshape=True)

        return lam

    def checkResiduum(self, verbose=False, force_only=True):

        # compute residual force and tangent stiffness
        self.assemble(force_only=force_only)

        normR = np.dot(self.R, self.R)

        # Add constraint violation in case we are using displacement control
        if self.hasConstraint:

            if self.useArcLength:
                # arc-length control
                dload = self.loadfactor - self.loadfactor_n
                self.g = self.arclength2 - self.alpha * dload*dload * self.P@self.P
                for node in self.nodes:
                    self.g -= node.getNormDeltaU2()

            else:
                # displacement control
                self.g = self.targetU - self.control_node.getDisp(self.control_dof)[0]

            normR += self.g*self.g

        else:
            self.g = 0.0

        normR = np.sqrt(normR)

        if verbose:
            #print('-')
            if self.hasConstraint:
                print(f"norm of the out-of-balance force: {normR:12.4e} with g={self.g:12.4e}")
            else:
                print(f"norm of the out-of-balance force: {normR:12.4e}")

        # convergence check
        return normR

    def resetForces(self):
        r"""
        Reset force vector to **all zeros**.
        """
        self.P = np.zeros(self.sdof)

    def resetDisplacements(self):
        r"""
        Reset displacement vector to **all zeros**.
        """
        self.sysU = np.zeros(self.sdof)

    def showKt(self, filename="", **kwargs):

        plt.figure()
        plt.spy(self.Kt, marker='.',mec='b',mfc='b', **kwargs)
        if filename:
            plt.savefig(filename)
        else:
            plt.show()

    def getNodalReactions(self, dofs=None, cut_off=1.0e-6):

        self.assemble(force_only=True)

        R = []

        # assemble loads
        for node in self.nodes:

            reaction = np.zeros(3)

            if 'ux' in node.dofs:
                reaction[0] = self.R[node.getIdx4DOFs(dofs=['ux'])]
            if 'uy' in node.dofs:
                reaction[1] = self.R[node.getIdx4DOFs(dofs=['uy'])]
            if 'rz' in node.dofs:
                reaction[2] = self.R[node.getIdx4DOFs(dofs=['rz'])]

            if np.linalg.norm(reaction) <= cut_off:
                reaction = np.zeros_like(reaction)

            R.append(reaction)

        return R

    def getNodalLoads(self, dofs=None, cut_off=1.0e-6):

        R = []

        # collect nodal loads
        for node in self.nodes:
            R.append(node.getLoad(dof_list=dofs, apply_load_factor=True))

        # Element Loop: assemble element forces and stiffness
        for element in self.elements:
            Pe = element.getLoad()             # Element State Update occurs here
            for (i,ndI) in enumerate(element.nodes):
                while not ndI.isLead():
                    ndI = ndI.getLead()

                node_idx = self.nodes.index(ndI)
                if isinstance(Pe[i], np.ndarray):
                    if dofs:
                        element_dofs = element.getDofs()  # dof list for this element
                        P_filtered = []
                        for dof_key in dofs:
                            if dof_key in element_dofs:
                                idx = element_dofs.index(dof_key)
                                P_filtered.append(Pe[i][idx])
                            else:
                                P_filtered.append(0.0)
                        R[node_idx] += self.loadfactor * np.array(P_filtered)
                    else:
                        R[node_idx] += self.loadfactor * Pe[i]

        # Lds = []
        # for node in self.nodes:
        #     force = R[node]
        #     if np.linalg.norm(force) <= cut_off:
        #         force = np.zeros_like(force)
        #     Lds.append(force)

        return R


    def initArcLength(self, load_increment=1., alpha=0.0, tolerance=1.0e-12):
        r"""
        This method may be implemented by a nonlinear solver
        """
        msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        raise NotImplementedError(msg)

    def stepArcLength(self, verbose=False):
        r"""
        This method may be implemented by a nonlinear solver
        """
        msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        raise NotImplementedError(msg)

    def startRecorder(self):
        self.record = True

    def pauseRecorder(self):
        self.record = False

    def stopRecorder(self):
        self.record = False

    def recordThisStep(self):
        r"""
        initiate writing of the current state to all existing recorders.
        (system, nodes, elements, materials - as supported by those objects)

        .. note::

           This method is automatically initiated by the solver
           for every converged solution.  No data will be recorded if a
           load step fails to converge to the target tolerance.

           Additional states can be recorded from the :py:class:`System` object.

        """
        # tell model to record
        if self.model_ptr:
            self.model_ptr.recordThisStep()





from .Recorder import *
from .Record import *

class NodeRecorder(Recorder):
    """

    """

    def __init__(self, **kwargs):
        super(NodeRecorder, self).__init__(**kwargs)

        if 'label' in kwargs:
            template = kwargs['label'] + ":{}"
        else:
            template = "node:{}"

        for var in self.nodevars:
            lbl = template.format(var)
            self.data[var] = Record(key=var, label=lbl)




"""
=====================================================
Finite elements based on a variational formulation.
=====================================================

Elements in this group share the following features:
* small deformation kinematics
* based on variational principles
* numeric integration of stiffness and internal force, thus, allowing for the use of inelastic material
"""

__all__ = (
    "Element",
    "LinearElement",
    "DrawElement",
    "Faces",
    "Face2D",
    "Face3D",
)

from .Faces import *
from .Face2D import *
from .Face3D import *
from .DrawElement import *
from .Element import *
from .LinearElement import *




"""
Example: frame04

Buckling of a building frame

modeled using a 2D frame element

.. list-table:: setting given parameters

    * - N  = 2
      - number of elements
    * - L  = 100.0
      - column length
    * - EA = 2000000.0
      - axial stiffness
    * - EI = 21000.0
      - flexural stiffness
    * - w  = 0.1
      - applied lateral load


"""
from ...examples.Example import *

from ...domain import *
from ...solver.NewtonRaphsonSolver import *
from ...elements.Frame2D import *
from ...materials.ElasticSection import *


class ExampleFrame04(Example):

    def docString(self):
        s = """
    Buckling of a building frame

    degrees of freedom:
        ux ... horizontal displacement
        uy ... vertical displacement
        rz ... rotation, theta
        
    Author: Peter Mackenzie-Helnwein 
    """
        return s

    def problem(self):
        # initialize a system model

        N  = 8     # number of elements

        B = 720.
        H = 720.

        E  = 29000.0
        A = 150.0
        I = 250.0

        w = 0.10
        load_at_nodes_only = False # set to True to apply equivalent nodal forces and moments

        Ph = 0.01      # additional horizontal load per floor
        Ph = 0.10      # additional horizontal load per floor
        Ph = 1.00      # additional horizontal load per floor
        Ph = 0.00      # additional horizontal load per floor

        # ========== setting global parameters ==============

        target_load_level = 33.
        max_steps = 10
        load_levels = np.linspace(0, target_load_level, max_steps)

        # ========= build your structural model =============

        model = System()
        model.setSolver(NewtonRaphsonSolver())

        x0 = 0.0
        x1 = B / 3
        x2 = 2 * B / 3
        x3 = B

        y0 = 0.0
        y1 = H / 4
        y2 = 2 * H / 4
        y3 = 3 * H / 4
        y4 = H

        X10 = Node(x0, y0)
        X11 = Node(x0, y1)
        X12 = Node(x0, y2)
        X13 = Node(x0, y3)
        X14 = Node(x0, y4)

        X20 = Node(x1, y0)
        X21 = Node(x1, y1)
        X22 = Node(x1, y2)
        X23 = Node(x1, y3)
        X24 = Node(x1, y4)

        X30 = Node(x2, y0)
        X31 = Node(x2, y1)
        X32 = Node(x2, y2)
        X33 = Node(x2, y3)
        X34 = Node(x2, y4)

        X40 = Node(x3, y0)
        X41 = Node(x3, y1)
        X42 = Node(x3, y2)
        X43 = Node(x3, y3)
        X44 = Node(x3, y4)

        model.addNode(X10,X11,X12,X13,X14)
        model.addNode(X20,X21,X22,X23,X24)
        model.addNode(X30,X31,X32,X33,X34)
        model.addNode(X40,X41,X42,X43,X44)

        # columns

        params = {'E': E, 'A': A, 'I': I}

        C11 = Frame2D(X10, X11, ElasticSection(params))
        C12 = Frame2D(X11, X12, ElasticSection(params))
        C13 = Frame2D(X12, X13, ElasticSection(params))
        C14 = Frame2D(X13, X14, ElasticSection(params))

        model.addElement(C11,C12,C13,C14)

        params = {'E': E, 'A': 2*A, 'I': 1.5*I}

        C21 = Frame2D(X20, X21, ElasticSection(params))
        C22 = Frame2D(X21, X22, ElasticSection(params))
        C23 = Frame2D(X22, X23, ElasticSection(params))
        C24 = Frame2D(X23, X24, ElasticSection(params))

        model.addElement(C21,C22,C23,C24)

        C31 = Frame2D(X30, X31, ElasticSection(params))
        C32 = Frame2D(X31, X32, ElasticSection(params))
        C33 = Frame2D(X32, X33, ElasticSection(params))
        C34 = Frame2D(X33, X34, ElasticSection(params))

        model.addElement(C31,C32,C33,C34)

        params = {'E': E, 'A': A, 'I': I}

        C41 = Frame2D(X40, X41, ElasticSection(params))
        C42 = Frame2D(X41, X42, ElasticSection(params))
        C43 = Frame2D(X42, X43, ElasticSection(params))
        C44 = Frame2D(X43, X44, ElasticSection(params))

        model.addElement(C41,C42,C43,C44)

        # floors

        params = {'E': E, 'A': A, 'I': 3*I}

        F11 = Frame2D(X11, X21, ElasticSection(params))
        F12 = Frame2D(X21, X31, ElasticSection(params))
        F13 = Frame2D(X31, X41, ElasticSection(params))

        model.addElement(F11,F12,F13)

        F21 = Frame2D(X12, X22, ElasticSection(params))
        F22 = Frame2D(X22, X32, ElasticSection(params))
        F23 = Frame2D(X32, X42, ElasticSection(params))

        model.addElement(F21,F22,F23)

        F31 = Frame2D(X13, X23, ElasticSection(params))
        F32 = Frame2D(X23, X33, ElasticSection(params))
        F33 = Frame2D(X33, X43, ElasticSection(params))

        model.addElement(F31,F32,F33)

        F41 = Frame2D(X14, X24, ElasticSection(params))
        F42 = Frame2D(X24, X34, ElasticSection(params))
        F43 = Frame2D(X34, X44, ElasticSection(params))

        model.addElement(F41,F42,F43)

        # fixities
        X10.fixDOF('ux','uy','rz')   # fixed
        X20.fixDOF('ux','uy','rz')   # fixed
        X30.fixDOF('ux','uy','rz')   # fixed
        X40.fixDOF('ux','uy','rz')   # fixed

        # reference load
        #Pcr = np.pi**2 * EI / L**2
        model.resetLoad()            # size load vector and initialize
        #model.addLoad(Xn, -Pcr, dof=0) # add a horizontal force (first dof only) ; remember C-style indexing: 0,1,...,(n-1)

        if load_at_nodes_only:

            # floor loading as nodal loads ...

            Pe = w * B/3
            Mi = w * (B/3)**2 /12

            X11.addLoad([-Pe/2., -Mi],['uy','rz'])
            X21.addLoad([-Pe/2.,  0.],['uy','rz'])
            X31.addLoad([-Pe/2.,  0.],['uy','rz'])
            X41.addLoad([-Pe/2.,  Mi],['uy','rz'])

            X12.addLoad([-Pe/2., -Mi],['uy','rz'])
            X22.addLoad([-Pe/2.,  0.],['uy','rz'])
            X32.addLoad([-Pe/2.,  0.],['uy','rz'])
            X42.addLoad([-Pe/2.,  Mi],['uy','rz'])

            X13.addLoad([-Pe/2., -Mi],['uy','rz'])
            X23.addLoad([-Pe/2.,  0.],['uy','rz'])
            X33.addLoad([-Pe/2.,  0.],['uy','rz'])
            X43.addLoad([-Pe/2.,  Mi],['uy','rz'])

            X14.addLoad([-Pe/2., -Mi],['uy','rz'])
            X24.addLoad([-Pe/2.,  0.],['uy','rz'])
            X34.addLoad([-Pe/2.,  0.],['uy','rz'])
            X44.addLoad([-Pe/2.,  Mi],['uy','rz'])

        else:

            # floor loading as distributed loads ...

            F11.setDistLoad(-w)
            F12.setDistLoad(-w)
            F13.setDistLoad(-w)

            F21.setDistLoad(-w)
            F22.setDistLoad(-w)
            F23.setDistLoad(-w)

            F31.setDistLoad(-w)
            F32.setDistLoad(-w)
            F33.setDistLoad(-w)

            F41.setDistLoad(-w)
            F42.setDistLoad(-w)
            F43.setDistLoad(-w)


        # wind load ...

        X11.addLoad([Ph],['ux'])   # horizontal load
        X12.addLoad([Ph],['ux'])   # horizontal load
        X13.addLoad([Ph],['ux'])   # horizontal load
        X14.addLoad([Ph/2],['ux']) # horizontal load


        # show model information
        print(model)

        print("\n==== perform the analysis ===\n")

        # * apply the load in multiple smaller load steps

        # set up data recorder
        model.initRecorder()
        model.trackStability(True)

        # initialize the analysis:
        model.resetDisplacements()   # set U to all zeros
        model.setLoadFactor(0.0)     # define a known equilibrium solution

        model.startRecorder()

        detKt   = []
        lambdas = []

        # solve for all load_levels
        for loadfactor in load_levels:

            # define node X2 as the controled node; downward direction is prescribed:
            model.setLoadFactor(loadfactor)
            model.solve(verbose=True)

            # stability check
            lambdas.append(model.loadfactor)
            detKt.append(model.solver.checkStability())

            # report results
            print('+')
            #model.report()

            print("\n=== next load level ===\n")


        #
        # ==== create some nice plots ===
        #

        model.report()

        print(lambdas)
        print(detKt)

        plt.plot(lambdas,detKt,'--*r')
        plt.grid(True)
        plt.xlabel('Load factor, $ \\lambda $')
        plt.ylabel("Stability index, $ {det}\: {\\bf K}_t $")
        plt.show()

        model.plot(factor=10.0, filename="frame4_deformed.png")

        model.beamValuePlot("F", filename="frame4_force.png")
        model.beamValuePlot("V", filename="frame4_shear.png")
        model.beamValuePlot("M", filename="frame4_moment.png")

        model.plotBucklingMode(factor=100., mode=0, filename="frame4_buckling_mode0.png")
        model.plotBucklingMode(factor=100., mode=1, filename="frame4_buckling_mode1.png")
        model.plotBucklingMode(factor=100., mode=2, filename="frame4_buckling_mode2.png")
        model.plotBucklingMode(factor=100., mode=3, filename="frame4_buckling_mode3.png")



import numpy as np

from ..Element import *
from ...domain.Node import *

class Truss(Element):
    """
    Representing a single truss element between 2 nodes.

    The element is using the following dofs:

    * :math:`u` (:code:`ux`) and :math:`v` (:code:`uy`) in a 2D model.
    * :math:`u` (:code:`ux`), :math:`v` (:code:`uy`) and :math:`w` (:code:`uz`) in a 3D model.

    .. list-table:: Internal variables

        * - self.nodes
          - nodes i and j (tuple)
        * - self.material
          - pointer to Material object
        * - self.force
          - internal force (float)
        * - self.Forces
          - nodal force vectors (list of np.arrays)
        * - self.Kt
          - tangent stiffness (list of np.arrays)

    """

    def __init__(self, nodei, nodej, material, label=None):
        super().__init__((nodei, nodej), material, label=label)
        self.element_type = DrawElement.LINE

        if not (self.material.materialType() == Material.SECTION1D
                or self.material.materialType() == Material.FIBER):
            msg = "Incompatible material type: need FIBER or SECTION1D"
            raise TypeError(msg)

        dim = nodei.getPos().size

        if dim == 3:
            dof_list = ('ux', 'uy', 'uz')
        elif dim == 2:
            dof_list = ('ux', 'uy')
        else:
            raise TypeError("spatial dimension of nodes must be 2 or 3")

        self._requestDofs(dof_list)

        self.L0       = np.linalg.norm(self.nodes[1].getPos() - self.nodes[0].getPos())
        self.force    = 0.0
        self.Forces   = [np.zeros(dim), np.zeros(dim)]
        self.Kt       = [[np.zeros((dim, dim)), np.zeros((dim, dim))],
                         [np.zeros((dim, dim)), np.zeros((dim, dim))]]

        """
        Compute internal state, nodal forces, and tangent stiffness for the current state of deformation.
        """
        X0 = self.getPos(0)
        X1 = self.getPos(1)

        # local coordinate system
        Lvec = X1 - X0
        L = np.linalg.norm(Lvec)   # element length (undeformed)
        Nvec = Lvec / L            # axial vector (unity)

        # get initial material stiffness
        self.material.setStrain({'xx':0.0})     # set strain to 0.0 for the initial stiffness
        EA   = self.material.getStiffness() * self.material.getArea()

        # nodal stiffness and element stiffness
        ke = (EA / L) * np.outer(Nvec, Nvec)
        self.Kt = [[ke, -ke], [-ke, ke]]

        # remember for force recovery
        self.L    = L
        self.Nvec = Nvec
        self.EA   = EA


    def __str__(self):
        area = self.material.getArea()
        stress = self.force / area
        strain = self.force / self.EA
        s = \
"""Truss: {} to {}:
    material properties: EA:{}  strain:{}   stress:{}  
    internal force: {}""".format( self.nodes[0].getID(), self.nodes[1].getID(),
                            self.EA, strain, stress, self.force)
        return s

    def __repr__(self):
        return "Truss({},{},{})".format( repr(self.nodes[0].getID()),
                                         repr(self.nodes[1].getID()),
                                         repr(self.material))

    def getAxialForce(self):
        """
        :returns: axial force in the truss. Positive value = tension.
        """
        self.updateState()
        return self.force

    def getInternalForce(self, variable=''):
        self.updateState()

        if variable.lower() == 'f' or variable.lower() == 'fx':
            # axial force
            s   = np.array([0.,1.])
            val = np.array([self.force, self.force])
            return (s,val)

        elif variable.lower() == 'eps' or variable.lower() == 'strain':
            # axial force
            s   = np.array([0.,1.])
            strain = self.force / self.EA
            val = np.array([strain,strain])
            return (s,val)

        else:
            return (np.empty([0]),np.empty([0]))

    def updateState(self):
        """
        Stress/force recovery
        """
        L    = self.L
        Nvec = self.Nvec
        EA   = self.EA

        U0 = self.getDisp(0)
        U1 = self.getDisp(1)

        # compute strain
        eps = Nvec @ (U1 - U0) / L

        # compute internal force
        self.force = EA * eps

        # nodal force vector
        Pe = self.force * Nvec
        self.Forces = [-Pe, Pe]




"""
=====================================================
Finite elements based on the direct stiffness method
=====================================================

Elements in this group share the following features:
* small deformation kinematics
* linear elastic material behavior
* based on the analytical solution of the governing equations
"""

__all__ = (
    "Truss",
    "Beam2D",
    "Frame2D",
)

from .Beam2D import *
from .Frame2D import *
from .Truss import *



import numpy as np

from ..Element import *
from ...materials.Material import Material
from ...domain.Node import *

class Truss(Element):
    """
    Representing a single truss element for finite deformation analysis between 2 nodes.

    The element is using the following dofs:

    * :math:`u` (:code:`ux`) and :math:`v` (:code:`uy`) in a 2D model.
    * :math:`u` (:code:`ux`), :math:`v` (:code:`uy`) and :math:`w` (:code:`uz`) in a 3D model.

    .. list-table:: Internal variables

        * - self.nodes
          - nodes i and j (tuple)
        * - self.material
          - pointer to Material object
        * - self.force
          - internal force (float)
        * - self.Forces
          - nodal force vectors (list of np.arrays)
        * - self.Kt
          - tangent stiffness (list of np.arrays)

    """

    def __init__(self, nodei, nodej, material, label=None):
        super().__init__((nodei, nodej), material, label=label)
        self.element_type = DrawElement.LINE

        if not (self.material.materialType() == Material.SECTION1D
                or self.material.materialType() == Material.FIBER):
            msg = "Incompatible material type: need FIBER or SECTION1D"
            raise TypeError(msg)

        dim = nodei.getPos().size

        if dim == 3:
            dof_list = ('ux', 'uy', 'uz')
        elif dim == 2:
            dof_list = ('ux', 'uy')
        else:
            raise TypeError("spatial dimension of nodes must be 2 or 3")

        self._requestDofs(dof_list)

        # local coordinate system
        X0 = self.getPos(0)
        X1 = self.getPos(1)

        Lvec = X1 - X0
        self.L0   = np.linalg.norm(Lvec)
        self.Nvec = Lvec / self.L0

        # initialize forces and stiffness
        self.force    = 0.0
        self.Forces   = [np.zeros(dim), np.zeros(dim)]
        self.Kt       = [[np.zeros((dim, dim)), np.zeros((dim, dim))],
                         [np.zeros((dim, dim)), np.zeros((dim, dim))]]


    def __str__(self):
        s = \
"""Truss: {} to {}:
    material properties: {}  strain:{}   stress:{}  
    internal force: {}""".format( self.nodes[0].getID(), self.nodes[1].getID(),
                            repr(self.material), self.material.getStrain(),
                            self.material.getStress(), self.force)
        return s

    def __repr__(self):
        return "NLTruss({},{},{})".format( repr(self.nodes[0].getID()),
                                         repr(self.nodes[1].getID()),
                                         repr(self.material))

    def getAxialForce(self):
        """
        :returns: axial force in the truss. Positive value = tension.
        """
        self.updateState()
        return self.force

    def getInternalForce(self, variable=''):
        self.updateState()

        if variable.lower() == 'f' or variable.lower() == 'fx':
            # axial force
            s   = np.array([0.,1.])
            val = np.array([self.force, self.force])
            return (s,val)

        elif variable.lower() == 'eps' or variable.lower() == 'strain':
            # axial force
            s   = np.array([0.,1.])
            strain = self.material.getStrain()
            val = np.array([strain,strain])
            return (s,val)

        else:
            return (np.empty([0]),np.empty([0]))

    def updateState(self):
        """
        Compute internal state, nodal forces, and tangent stiffness for the current state of deformation.
        """
        U0 = self.getDisp(0)
        X0 = self.getPos(0)
        U1 = self.getDisp(1)
        X1 = self.getPos(1)

        # local coordinate system
        lvec = (X1 + U1) - (X0 + U0)
        ell = np.linalg.norm(lvec)
        nvec = lvec / ell

        # kinematics: Henky strain
        eps = np.log( ell/self.L0 )

        # constitutive behavior
        self.material.setStrain({'xx':eps})
        stress = self.material.getStress()
        sig    = stress['xx']

        # stress resultant
        area   = self.material.getArea()
        self.force = sig * area

        # nodal forces
        Pe = self.force * nvec
        self.Forces = [-Pe, Pe]

        # nodal and element stiffness matrix
        Et = self.material.getStiffness()
        n_outer_n = np.outer(nvec, nvec)
        ke = (Et * area / ell) * n_outer_n + self.force / ell * (np.eye(len(nvec)) - n_outer_n)
        self.Kt = [[ke, -ke], [-ke, ke]]



from .Constraint import Constraint

class TieNodes(Constraint):
    """
    Use this to tie some or all dofs of two or more nodes.

    The program will select a **lead** node and assign all remaining nodes
    as **followers**.

    .. note::

        Most textbooks refer to such nodes as "master" and "slave" nodes,
        which we avoid on purpose due to their historic use in slavery
        and reappearance in modern day racism.

        Please view this is an apology for mistakes of the past
        and help us eradicate racism along the way.


    """

    def __init__(self, *nodes, dofs=('all',), **kwargs):
        """
        :param nodes: two or mode :py:class:`Node` objects
        :param dofs:  (optional) a list of dof-codes to be tied
                The default code :py:code:`"all"` will tie all dofs shared by the nodes.

        """
        params = {}
        params['dofs'] = dofs

        if 'all' in dofs:
            self.neqns = len(nodes[0].ndofs)  # needs fancier logic for multi-node ties
        else:
            self.neqns  = len(dofs)

        super(TieNodes, self).__init__(nodes, params, **kwargs)



import sys
import numpy as np

from .ShapeFunctions import ShapeFunctions

class TriangleShapes(ShapeFunctions):
    r"""
    Shape functions or their `*n-th` mixed derivative
    for the unit-triangle (two-dimensional domain :math:`(s,t)\in[0,+1]\times[0,+1]` with :math:`s+t\le +1`).

    Derivatives are with respect to the normalized coordinate and need to be scaled accordingly.

    Triangle shape functions are commonly formulated in triangle coordinates :math:`(\xi_1,\xi_2,\xi_3)`
    where :math:`\xi_1+\xi_2+\xi_3=1`.
    This implementation uses :math:`s\to\xi_1` and :math:`t\to\xi_2`.
    The third component is implicitly computed as :math:`\xi_3=1-\xi_1-\xi_2`.
    The derivatives are returned as

    .. math::
       \begin{array}{rcl}
          \frac{\partial }{\partial s} &=& \frac{\partial }{\partial \xi_1} - \frac{\partial }{\partial \xi_3} \\[2ex]
          \frac{\partial }{\partial t} &=& \frac{\partial }{\partial \xi_2} - \frac{\partial }{\partial \xi_3}
       \end{array}

    """

    def __init__(self):
        super(TriangleShapes, self).__init__(ShapeFunctions.TRIANGLE)

    def shape(self, order, s, t, n=(0,0)):
        '''

        :param order: polynomial order of the shape functions. Options: 0,1,2,
        :param s:    first coordinate of point of interest on interval [-1,1]
        :type s:  float or np.array
        :param t:   second coordinate location on interval [-1,1]
        :type t:  float or np.array
        :param n:     return n-th derivative of the requested shape function.
        :type n:  tuple or list

            Use :code:`n=(1,0)` for first derivative with respect to the first variable.
            Use :code:`n=(2,1)` for second derivative with respect to the first variable and
            first with respect to the second variable, i.e., a mixed derivative.

        :return: PHI: a list of shape function values
        '''
        ns, nt = n

        u = 1. - s - t

        if (order == 0):
        #
        #   constant shape function
        #
            if (n == 0):
                PHI = np.array([1.0])
            else:
                PHI = np.array([0.0])

        elif (order == 1):
        #
        #   linear shape functions
        #
            if (ns == 0 and nt == 0):
                PHI = np.array([ u, s, t ])
            elif (ns == 1 and nt == 0):
                PHI = np.array([ -1., 1., 0. ])
            elif (ns == 0 and nt == 1):
                PHI = np.array([ -1., 0., 1. ])
            else:
                PHI = np.array([0.0, 0.0, 0.0])

        elif (order == 2):
        #
        #   quadratic shape functions
        #
            if (ns == 0 and nt == 0):
                PHI = np.array([ u*(2*u-1), s*(2*s-1), t*(2*t-1), 4*u*s, 4*s*t, 4*t*u ])
            elif (ns == 1 and nt == 0):
                PHI = np.array([ -(4*u-1), (4*s-1), 0.0, 4*(u-s), 4*t, -4*t ])
            elif (ns == 0 and nt == 1):
                PHI = np.array([ -(4*u-1), 0.0, (4*t-1), -4*s, 4*s, 4*(u-t) ])
            elif (ns == 2 and nt == 0):
                PHI = np.array([ 4.0, 4.0, 0.0,-8.0, 0.0, 0.0 ])
            elif (ns == 1 and nt == 1):
                PHI = np.array([ 4.0, 0.0, 0.0, -4.0, 4.0, -4.0 ])
            elif (ns == 0 and nt == 2):
                PHI = np.array([ 4.0, 0.0, 4.0, 0.0, 0.0, -8.0 ])
            else:
                PHI = np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0])

        else:
            msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
            raise NotImplementedError(msg)

        return PHI



import numpy as np
import matplotlib.pyplot as plt


class Material():
    r"""
    abstract class: representing a generic material

    """

    """
    defining class parameters
    
    Add a new one for every general type of material.
    This will be used by elements to test for compatible materials
    """
    UNKNOWN     = 0x000000
    DEFAULT     = 0x000001
    FIBER       = 0x000002
    SECTION1D   = 0x000004  # integrated section (beams and frames)
    SECTION2D   = 0x000008  # through the thickness integration (plates and shells)
    PLANESTRESS = 0x000010
    PLANESTRAIN = 0x000020
    CONTINUUM   = 0x000040

    ELASTIC     = 0x010000
    PLASTIC     = 0x020000
    CREEP       = 0x040000
    HARDENING   = 0x080000
    DIFFUSION   = 0x100000


    def __init__(self, params={'E':1.0, 'A':1.0, 'nu':0.0, 'fy':1.0e30}):
        """

        :param params:
        """
        self._type = self.UNKNOWN

        self.parameters = params

        # make sure all necessary parameters exist
        if 'E' not in self.parameters:
            self.parameters['E']  = 1.0
        if 'A' not in self.parameters:
            self.parameters['A']  = 1.0
        if 'nu' not in self.parameters:
            self.parameters['nu'] = 0.0
        if 'fy' not in self.parameters:
            self.parameters['fy'] = 1.0e30

        self.sig = 0.0
        self.Et  = self.parameters['E']

    def __str__(self):
        s = "{}(Material)({})".format(self.__class__.__name__, self.parameters)
        return s

    def __repr__(self):
        return str(self)

    def materialType(self):
        return self._type

    def isMaterialType(self, type):
        return ( self._type == type )

    def getStress(self):
        r"""
        request axial stress

        :return: sigma
        """
        return self.stress

    def getStiffness(self):
        r"""
        request axial stiffness

        :return: Et ... tangent material stiffness
        """
        return self.Et

    def setStrain(self, eps):
        r"""
        update state for a user provided axial strain value

        :param eps:  strain or strain tensor
        """
        # update stress state
        self.strain = eps
        self.updateState()

    def updateState(self):
        raise NotImplementedError(self.__class__.__name__ + '.updateState() needs to be overloaded')

    def getStrain(self):
        return self.strain

    def converged(self):
        raise NotImplementedError(self.__class__.__name__ + '.converged() needs to be overloaded')


if __name__ == "__main__":
    # testing the Node class
    mat = Material(params={'E':100.0, 'nu':0.0, 'fy':1.0})

    eps = 0.02 * np.sin( np.linspace(0, 2.*np.pi, 100) )

    sig = []
    Et = []
    for strain in eps:
        # update material strain state
        mat.setStrain(strain)
        #collect stress response
        sig.append(mat.getStress())
        #collect material tangent stiffness response
        Et.append(mat.getStiffness())

    fig, (ax1,ax2) = plt.subplots(2,1)

    ax1.plot(eps, sig, '-r', label='stress')
    ax1.grid(True)
    ax1.set_xlabel('strain $\\varepsilon$')
    ax1.set_ylabel('stress $\\sigma$')
    ax1.legend()

    ax2.plot(eps, Et, '-.b', label='tangent modulus')
    ax2.grid(True)
    ax2.set_xlabel('strain $\\varepsilon$')
    ax2.set_ylabel('tangent modulus $E_t$')
    ax2.legend()

    plt.show()



r"""
===================================================================
Patch test for the finite deformation 9-node quadrilateral
===================================================================
Full bi-quadratic interpolation

.. code::

             1
          x     y
      x^2   x*y   y^2
        x^2*y x*y^2
          x^2*y^2


"""

import numpy as np
from copy import deepcopy

from ..Element import *
from ...domain.Node import *
from ...utilities import QuadIntegration, QuadShapes, GPdataType

class Quad9(Element):
    r"""
    class: representing a plane 9-node quadrilateral

    This element works as 2D plate, loaded in-plane, and as a 3D membrane element.

    * For 2D plate behavior, define nodes as two-dimensional nodes
    * For 3D membrane behavior, define nodes as three-dimensional nodes
    """

    def __init__(self, node0, node1, node2, node3, node4, node5, node6, node7, node8, material, label=None):
        super(Quad9, self).__init__((node0, node1, node2, node3, node4, node5, node6, node7, node8), material, label=label)
        self.element_type = DrawElement.QUAD
        self.createFaces()

        ndim = node0.getPos().size

        if ndim == 3:
            dof_list = ('ux','uy','uz')
            ndof = 3
        elif ndim == 2:
            dof_list = ('ux','uy')
            ndof = 2
        else:
            raise TypeError("dimension of nodes must be 2 or 3")

        self._requestDofs(dof_list)

        self.distributed_load = [0.0, 0.0, 0.0, 0.0]   # face loads along the perimeter of the element
        self.force    = 0.0
        self.Forces   = [ np.zeros(ndof) for k in range(len(self.nodes)) ]
        self.Kt       = [ [ np.zeros((ndof,ndof)) for i in range(len(self.nodes)) ] for j in range(len(self.nodes)) ]
        self.ndof = ndof

        #self.material  = []   # one material object per integration point
        #self.stress    = []   # hold gauss-point stress (1st Piola-Kirchhoff stress)
        #self.strain    = []   # hold gauss-point strain
        #self.J         = []   # jacobian at integration point

        #self.Grad      = []   # derivative of shape functions with respect to global coords

        X  = np.array([ node.getPos() for node in self.nodes ])

        # initialization step
        self.integrator = QuadIntegration(order=4)
        self.xis, self.wis = self.integrator.parameters()

        self.gpData = [ GPdataType() for i  in range(len(self.xis)) ]

        gpt = 0
        gp2nd_map = []

        self.interpolation = QuadShapes()

        for xi, wi, gpData in zip(self.xis, self.wis, self.gpData):

            gpData.material = deepcopy(material)

            shape   = self.interpolation.shape(  # requesting shape function array
                order=2,          # polynomial order per direction: quadratic
                s=xi[0], t=xi[1], # local coordinates for current position
                n=(0,0))          # no derivative with respect to (s,t)

            dshape1 = self.interpolation.shape(  # d shape function / d xi1
                order=2,          # polynomial order per direction: quadratic
                s=xi[0], t=xi[1], # local coordinates for current position
                n=(1,0))          # first derivative with respect to s

            dshape2 = self.interpolation.shape(  # d shape function / d xi2
                order=2,          # polynomial order per direction: quadratic
                s=xi[0], t=xi[1], # local coordinates for current position
                n=(0,1))          # first derivative with respect to t

            Grad = np.vstack((dshape1, dshape2))

            # covariant base vectors (reference system)
            Xgp   = np.zeros(ndim)
            base1 = np.zeros(ndim)
            base2 = np.zeros(ndim)
            for N, dN1, dN2, node in zip(shape, dshape1, dshape2, self.nodes):
                Xgp   +=  N  * node.getPos()
                base1 += dN1 * node.getPos()
                base2 += dN2 * node.getPos()

            gpData.X = Xgp
            gcov = np.vstack((base1, base2))
            gpData.base = gcov

            # metric (reference system)
            GIJ = gcov @ gcov.T
            gpData.state['GIJ'] = GIJ

            # dual base vectors (reference system)
            gpData.dual_base = np.linalg.inv(GIJ) @ gcov

            gpData.J = np.sqrt(np.linalg.det(GIJ))



            # reference configuration
            # -------------------------

            #
            # $ D\Phi_0 $
            #
            DPhi0 = (Grad @ X).T
            #self.J.append( np.linalg.det(DPhi0) )  # material model already includes thickness

            # dual base (contra-variant)
            gpData.Grad = np.linalg.inv(DPhi0).T @ Grad

            # populate gauss-point to nodes map
            raw_map= self.interpolation.shape(  # requesting shape function array
                order=1,  # polynomial order per direction: quadratic
                s=xi[0], t=xi[1],  # local coordinates for current position
                n=(0, 0))  # no derivative with respect to (s,t)
            map = raw_map * wi * gpData.J
            gp2nd_map.append(map)
            gpt += 1

        self.ngpts      = gpt                    # number of gauss points
        self._gp2nd_map = np.array(gp2nd_map).T  # gauss-point to nodes map array


    def __str__(self):
        s = super(Quad9, self).__str__()
        for igpt, gpData in enumerate(self.gpData):
            material = gpData.material
            s += "\n    strain ({}): xx={xx:.3e} yy={yy:.3e} xy={xy:.3e} zz={zz:.3e}".format(igpt,**material.getStrain())
            s += "\n    stress ({}): xx={xx:.3e} yy={yy:.3e} xy={xy:.3e} zz={zz:.3e}".format(igpt,**material.getStress())
        if np.array(self.distributed_load).any():
            s += "\n    element forces added to node:"
            for i, P in enumerate(self.Loads):
                Pi = np.array(P) * self.loadfactor
                s += "\n        {}: {}".format(self.nodes[i].getID(), Pi)
        return s

    def setSurfaceLoad(self, face, pn, ps=0):
        r"""
        .. list-table::
            :header-rows: 1

            * - face ID
              - nodes defining that face
            * - 0
              - :py:obj:`node 0` to :py:obj:`node 4` to :py:obj:`node 1`
            * - 1
              - :py:obj:`node 1` to :py:obj:`node 5` to :py:obj:`node 2`
            * - 2
              - :py:obj:`node 2` to :py:obj:`node 6` to :py:obj:`node 3`
            * - 3
              - :py:obj:`node 3` to :py:obj:`node 7` to :py:obj:`node 0`


        :param face: face ID for the laoded face
        :param pn: magnitude of distributed normal load per area. Tension on a surface is positive.
        :param ps: magnitude of distributed shear load per area. Positive direction is defined as shown in the above table.
        """
        if face >= 0 and face <= 3:
            self.faces[face].setLoad(pn, ps)

    def resetLoads(self):
        super(Quad9, self).resetLoads()

    def updateState(self):

        # initialization step
        ndim = self.nodes[0].getPos().size

        nnds = len(self.nodes)
        ndof = self.ndof       # mechanical element

        self.Forces = [ np.zeros(ndof) for k in range(nnds) ]
        Kt = [ [ np.zeros((ndof,ndof)) for i in range(nnds) ] for j in range(nnds) ]

        # create array of deformed nodal coordinates
        xt = np.array([ node.getDeformedPos() for node in self.nodes ])

        for xi, wi, gpData in zip(self.xis, self.wis, self.gpData):

            # grab pre-computed quantities
            Grad = gpData.Grad
            wi  *= gpData.J

            # spatial configuration
            # -------------------------

            # deformation gradient
            F = (Grad @ xt).T

            # strain
            eps = 0.5 * ( F.T @ F - np.eye(ndim) )

            # update the material state
            strain = {'xx':eps[0,0], 'yy':eps[1,1], 'xy':eps[0,1]+eps[1,0]}
            gpData.state['strain'] = strain
            gpData.material.setStrain(strain)

            # 2nd Piola-Kirchhoff stress
            stress =gpData.material.getStress()

            S = np.array( [[stress['xx'],stress['xy']],[stress['xy'],stress['yy']]] )

            # 1st Piola-Kirchhoff stress
            P = F @ S

            # store stress for reporting
            stress = {'xx':P[0,0], 'xy':P[0,1], 'yx':P[1,0], 'yy':P[1,1]}
            gpData.state['stress'] = stress

            # compute kinematic matrices
            BI = [ np.array([ Grad[0,K]*F[:,0],                       # the XX component
                              Grad[1,K]*F[:,1],                       # the YY component
                              Grad[0,K]*F[:,1] + Grad[1,K]*F[:,0] ])  # the XY component is XY + YX ("gammaXY = 2 epsXY")
                   for K in range(nnds) ]

            # internal forces
            for i, force in enumerate(self.Forces):
                # self.Forces[i] += P @ Grad[:,i] * wi  # wi already includes J
                force += P @ Grad[:,i] * wi  # wi already includes J

            # tangent stiffness
            Ct = gpData.material.getStiffness() * wi

            One = np.eye(2, dtype=np.float64)
            for I, Bi in enumerate(BI):
                Ti   = Grad[:,I] @ S
                GCti = Bi.T @ Ct
                for J, Bj in enumerate(BI):
                    GIJ = Ti @ Grad[:,J] * wi
                    Kt[I][J] += GCti @ Bj + GIJ * One

            # on to the next integration point

        self.Kt = Kt

    def computeSurfaceLoads(self):
        r"""
        compute surface loads using faces

        This method should be called during :py:meth:`updateState()` by every
        element supporting surface loads

        Nodes are mapped between the element and each Face2D as follows

        .. list-table:: Mapping element node IDs to face node IDs
            :header-rows: 1

            * - Face ID
              - Face nodes (0,1,2)
            * - 0
              - (0,4,1)
            * - 0
              - (1,5,2)
            * - 0
              - (2,6,3)
            * - 0
              - (3,7,0)


        """
        self.Loads = [ np.zeros_like(self.Forces[I]) for I in range(len(self.nodes)) ]

        for I, face in enumerate(self.faces):
            loads = face.computeNodalForces()

            # indexing
            J = I+1
            if J>3:
                J -= 4

            # add to element load vectors
            self.Loads[I] += loads[0]
            self.Loads[J] += loads[2]
            if loads.shape[0]>2:
                numNodes = len(self.nodes)
                if numNodes == 8 or numNodes == 9:
                    K = I + 4
                else:
                    msg = "Force data provided from Face2D inconsistent with element data"
                    raise TypeError(msg)

                self.Loads[K] += loads[1]

    def getStress(self):
        stress = [ data.state['stress'] for data in self.gpData ]
        return stress

    def mapGaussPoints(self, var):
        r"""
        Initiate mapping of Gauss-point values to nodes.
        This method is an internal method and should not be called by the user.
        Calling that method explicitly will cause faulty nodal values.

        :param var: variable code for a variable to be mapped from Gauss-points to nodes
        """
        stresses = ('sxx','syy','szz','sxy','syz','szx')
        membrane = ('nxx','nyy','nxy')
        strains  = ('epsxx','epsyy','epszz','epsxy','epsyz','epszx')

        values = np.zeros( self.ngpts )

        if var.lower() in stresses:
            key = var[1:3].lower()
            values = []
            for gpdata in self.gpData:   # gauss-point loop
                if key in gpdata.state['stress']:
                    thickness = gpdata.material.getThickness()
                    values.append(gpdata.state['stress'][key]/thickness)
                else:
                    values.append(0.0)

        if var.lower() in membrane:
            key = var[1:3].lower()
            values = []
            for gpdata in self.gpData:   # gauss-point loop
                if key in gpdata.state['stress']:
                    values.append(gpdata.state['stress'][key])
                else:
                    values.append(0.0)

        if var.lower() in strains:
            key = var[1:3].lower()
            values = []
            for gpdata in self.gpData:   # gauss-point loop
                if key in gpdata.state['strain']:
                    values.append(gpdata.state['strain'][key])
                else:
                    values.append(0.0)

        values = np.array(values)

        for node, map in zip(self.nodes, self._gp2nd_map):
            wndi   = np.sum(map)
            val_wi = map @ values
            node._addToMap(wndi, val_wi)


import numpy as np

from ..Element import *
from ...materials.Material import *
from ...domain.Node import *

class BeamSolidLink(Element):
    r"""
    class: linking a node from a 2d model (follower) to a frame node (lead)

    **Assumptions**

    * The frame and the 2d-plate are in the X-Y-plane.
    * Small displacements and moderate rotations
    * Navier's assumptions (plane sections)

    The element is using dofs
    * :math:`u` (:code:`ux`), :math:`v` (:code:`uy`) and :math:`\theta` (:code:`rz`) on the frame node
    * :math:`u` (:code:`ux`) and :math:`v` on the 2d-plate node
    * one unique lagrange multiplier per link (no generic variable name)

    .. list-table:: Internal variables

        * - self.nodes
          - nodes i and j (tuple)
        * - self.material
          - pointer to Material object
        * - self.force
          - internal force (list of arrays for axial force, _f_, shear, _V_, and moment, _M_)
        * - self.Forces
          - nodal force vectors (list of np.arrays)
        * - self.Kt
          - tangent stiffness (list of np.arrays)

    """

    COUNT = 0

    def __init__(self, frame_node, plate_node, label=""):
        """
        :param frame_node: pointer to the lead node on the frame model
        :param plate_node: pointer to the follower node on the 2d-plate
        """
        super().__init__((frame_node, plate_node), None, label=label)
        self.element_type = DrawElement.LINE

        dim = plate_node.getPos().size
        if dim != 2:
            raise TypeError("spatial dimension of nodes must be 2")

        # genrate a unique Lagrange Multiplier
        lm_name = self.getUniqueLMName()

        dof_list_lead = ('ux', 'uy', 'rz')
        dof_idx = frame_node.request(dof_list_lead, self)
        self.dof_idx[frame_node] = dof_idx  # marked for removal

        dof_list_follow = ('ux', 'uy', lm_name)
        dof_idx = plate_node.request(dof_list_follow, self)
        self.dof_idx[plate_node] = dof_idx  # marked for removal

        self._dof_list = (dof_list_lead, dof_list_follow)

        ndof = 3

        # initialize element load to zero
        self.distributed_load = 0.0

        self.L0       = np.linalg.norm(self.nodes[1].getPos() - self.nodes[0].getPos())
        self.force    = 0.0
        self.Forces   = [np.zeros(ndof), np.zeros(ndof)]
        self.Kt       = [[np.zeros((ndof, ndof)), np.zeros((ndof, ndof))],
                         [np.zeros((ndof, ndof)), np.zeros((ndof, ndof))]]
        self.internal_forces = {'fi':0.0, 'Vi':0.0, 'Mi':0.0, 'fj':0.0, 'Vj':0.0, 'Mj':0.0}

    def __str__(self):
        s = super(BeamSolidLink, self).__str__()
        if not 'Pw' in self.internal_forces:
            self.internal_forces['Pw'] = 0.0
        if not 'Mw' in self.internal_forces:
            self.internal_forces['Mw'] = 0.0
        s += "\n    internal forces: f0={fi:.2f} V0={Vi:.2f} M0={Mi:.2f} fl={fj:.2f} Vl={Vj:.2f} Ml={Mj:.2f} Pw={Pw:.2f} Mw={Mw:.2f}".format(**self.internal_forces)
        return s



    def getDofs(self, node=None):
        r"""
        returns the dof-codes for this element in a list
        """
        return self._dof_list

    def getUniqueLMName(self):
        name = "LM_BS_{BeamSolidLink.COUNT:d}"
        BeamSolidLink.COUNT += 1
        return name

    def getInternalForce(self, variable=''):
        """
        computes vectors of normalized locations (**s**: ndarray) for which
        values (**val**: ndarray) are provided.
        Values of **s** are normalized to the interval :math:`[0,1]`.

        :returns: tuple (s, val)
        """
        self.updateState()

        Xi = self.nodes[0].getPos()
        Xj = self.nodes[1].getPos()
        Nvec = Xj - Xi
        L = np.linalg.norm(Nvec)

        if variable.lower() == 'm' or variable.lower() == 'mz':
            # bending moment (in plane)
            if self.distributed_load:
                s   = np.linspace(0,1,10)
            else:
                s   = np.array([0.,1.])

            Ml  = self.internal_forces['Mi']
            Ml += self.internal_forces['Mw']
            Mr  = self.internal_forces['Mj']
            Mr += self.internal_forces['Mw']

            #val = np.array([Ml, Mr])
            val = Ml * (1. - s) + Mr * s - 0.5 * (self.distributed_load * self.loadfactor) * L*L * s * (1 - s)

        elif variable.lower() == 'v' or variable.lower() == 'vy':
            # transverse shear (in-plane)
            if self.distributed_load:
                s   = np.linspace(0,1,10)
            else:
                s   = np.array([0.,1.])

            Vl  = self.internal_forces['Vi']
            Vl -= self.internal_forces['Pw']
            #Vr  = self.internal_forces['Vj']
            #Vr += self.internal_forces['Pw']

            val = Vl + s * (self.distributed_load * self.loadfactor) * L

        elif variable.lower() == 'f' or variable.lower() == 'fx':
            # transverse shear (in-plane)
            Fl = self.internal_forces['fi']
            Fr = self.internal_forces['fj']
            s   = np.array([0.,1.])
            val = np.array([Fl, Fr])

        else:
            s   = np.array([0.,1.])
            val = np.zeros_like(s)

        return (s,val)

    def updateState(self):

        Xi = self.getPos(0)
        Xj = self.getPos(1)

        # 0 ... ux
        # 1 ... uy
        # 2 ... theta
        dispi = self.getDisp(0)
        Ui = dispi[:2]
        dispj = self.getDisp(1)
        Uj = dispj[:2]

        # compute local coordinate system
        Nvec = Xj - Xi
        L = np.linalg.norm(Nvec)
        Nvec /= L
        Svec = np.array([[0,-1],[1,0]]) @ Nvec

        nvec = (Xj + Uj) - (Xi + Ui)
        ell = np.linalg.norm(nvec)
        nvec /= ell
        svec = np.array([[0,-1],[1,0]]) @ nvec

        # - compute axial strain
        #strain = 0.5 * ((ell/L)**2 - 1.)                    # finite deformation
        #psi = Svec @ (Uj - Ui) / L                          # von Karman
        #strain = Nvec @ (Uj - Ui) / L + 0.5 * (psi * psi)   # von Karman
        strain = Nvec @ (Uj - Ui) / L                        # P-delta

        # - compute curvature
        """
        Treat flexure as linear elastic for now.
        This is achieved by feeding the material curvature==0
        and, thus, receiving the initial flexural tangent stiffness.
        The moment needs to be computed by alternative means.
        """
        #vi     = dispi[1]    # the transformation was performed when we received Ui
        #vj     = dispj[1]    # the transformation was performed when we received Uj

        vi = Ui @ Svec  # transformation not yet implemented
        vj = Uj @ Svec  # transformation not yet implemented
        thetai = dispi[2]
        thetaj = dispj[2]

        curvature = 0.0

        # - update material set strain
        self.material.setStrain({'axial':strain, 'flexure':curvature})

        # - force and moment from the material
        stress = self.material.getStress()
        if 'axial' in stress:
            self.force = stress['axial']
        else:
            self.force = 0.0
        if 'flexure' in stress:
            self.moment = stress['flexure']
        else:
            self.moment = 0.0

        # get tangent moduli from the material
        Et = self.material.getStiffness()
        if 'ax-ax' in Et:
            EA = Et['ax-ax']
        else:
            EA = 0.0
        if 'flx-flx' in Et:
            EI = Et['flx-flx']
        else:
            EI = 0.0
        if ('ax-flx' in Et and Et['ax-flx']) or ('flx-ax' in Et and Et['flx-ax']):
            print('coupling of axial and flexural behavior is ignored by this element')

        # ** axial portion

        # - compute nodal tangent stiffness for axial
        n_tensor_n = np.outer(Nvec, Nvec)                                     # P-delta
        ke  = (EA / L) * n_tensor_n                                           # P-delta

        # ** bending portion (P-delta)

        if self.use_p_delta and self.force < -EI / L**2 / 1000000:
            """
            use exact solution for compressive members
            """

            kappa  = np.sqrt(-self.force / EI) * L
            kappa2 = kappa / 2

            cot  = 1./np.tan(kappa)
            csc  = 1./np.sin(kappa)
            cot2 = 1./np.tan(kappa2)
            sec2 = 1./np.cos(kappa2)

            EIfact = EI / (L*L*L * (2. - kappa*cot2))
            kfu  = EIfact * kappa**3 * cot2
            kft  = EIfact * kappa**2 * L
            kmu  = kft

            EIfact = EI * kappa * csc / (L*( kappa*csc - sec2**2))
            kmt  = EIfact * (kappa*cot - 1.)
            kmth = EIfact * (1. - kappa*csc)

        elif self.use_p_delta and self.force > EI / L**2 / 1000000:
            """
            use exact solution for tensile members 
            """
            kappa  = np.sqrt(self.force / EI) * L

            coshk = np.cosh(kappa)
            sinhk = np.sinh(kappa)

            EIfact = EI / (L*L*L*(2. - 2.*coshk + kappa*sinhk))

            kfu = EIfact * kappa**3 * sinhk
            kft = EIfact * kappa**2 * L * (coshk - 1.)
            kmu =  kft
            kmt =  EIfact * kappa * L*L * (kappa*coshk - sinhk)
            kmth = EIfact * kappa * L*L * (sinhk - kappa)

        else:
            """
            use the linear stiffness to avoid random numeric errors
            """

            kfu = 12 * EI / L**3
            kft =  6 * EI / L**2
            kmu =  kft
            kmt =  4 * EI / L
            kmth = kmt / 2

        #s_tensor_s = np.outer(svec, svec)

        n_tensor_n = np.outer(Nvec, Nvec)
        s_tensor_s = np.outer(Svec, Svec)
        KtII = np.block( [[ kfu * s_tensor_s,  kft * Svec[:,np.newaxis]], [ kmu * Svec, kmt ]] )
        KtIJ = np.block( [[-kfu * s_tensor_s,  kft * Svec[:,np.newaxis]], [-kmu * Svec, kmth]] )
        KtJI = np.block( [[-kfu * s_tensor_s, -kft * Svec[:,np.newaxis]], [ kmu * Svec, kmth]] )
        KtJJ = np.block( [[ kfu * s_tensor_s, -kft * Svec[:,np.newaxis]], [-kmu * Svec, kmt ]] )

        Vi =  kfu * (vi - vj) + kft * (thetaj + thetai)
        Vj = -kfu * (vi - vj) - kft * (thetaj + thetai)

        Mi = kmu * (vi - vj) + kmt * thetai + kmth * thetaj
        Mj = kmu * (vi - vj) + kmth * thetai + kmt * thetaj

        Fi = Vi * Svec - self.force * (Nvec)                  # P-delta
        Fj = Vj * Svec + self.force * (Nvec)                  # P-delta

        # ** combine the effects

        # - add nodal forces for axial
        KtII[0:2,0:2] +=  ke
        KtIJ[0:2,0:2] += -ke
        KtJI[0:2,0:2] += -ke
        KtJJ[0:2,0:2] +=  ke

        # ** build element load vector and element tangent stiffness

        # .. internal force
        self.Fi = np.r_[Fi, Mi]
        self.Fj = np.r_[Fj, Mj]
        self.Forces = [self.Fi, self.Fj]

        # .. tangent stiffness
        self.Kt = [[KtII, KtIJ],[KtJI, KtJJ]]

        # internal forces at nodes
        if 'Pw' in self.internal_forces:
            Pw = self.internal_forces['Pw']
        else:
            Pw = 0.0

        if 'Mw' in self.internal_forces:
            Mw = self.internal_forces['Mw']
        else:
            Mw = 0.0
        self.internal_forces = {'fi':self.force, 'Vi': Vi, 'Mi':-Mi,
                                'fj':self.force, 'Vj':-Vj, 'Mj': Mj,
                                'Pw':Pw, 'Mw':Mw}



__all__ = (
    "Record",
    "Recorder",
    "ModelRecorder",
    "NodeRecorder",
    "ElementRecorder",
    "MaterialRecorder",
)

from .Record import *
from .Recorder import *
from .ModelRecorder import *
from .NodeRecorder import *
from .ElementRecorder import *
from .MaterialRecorder import *



"""
=====================================================
Finite elements based on a variational formulation.
=====================================================

Elements in this group share the following features:
* large or finite deformation kinematics
* based on variational principles
* numeric integration of stiffness and internal force, thus, allowing for the use of inelastic material
"""

__all__ = (
    "Triangle",
    "Triangle6",
    "Quad",
    "Quad8",
    "Quad9",
    "Truss",
    "Frame2D",
)

from .Frame2D import *
from .Quad import *
from .Quad8 import *
from .Quad9 import *
from .Triangle import *
from .Triangle6 import *
from .Truss import *




__all__ = (
    "Mesher",
    "CurveMesher",
    "PatchMesher",
    "TriPatchMesher",
    "SolidMesher",
)

from .CurveMesher import *
from .PatchMesher import *
from .TriPatchMesher import *
from .SolidMesher import *



import numpy as np

from .Frame2dTransformation import *

class Beam2dTransformation(Frame2dTransformation):
    r"""
    This is an alias for :py:meth:`Frame2dTransformation`.
    See documentation for the :py:meth:`Frame2dTransformation` class for further detail.

    :param dir1: in-plane vector defining the local x-direction
    :param dir2: in-plane vector used to define the local y-direction
    :param angle: rotation angle in degrees
    """

    def __init__(self, dir1=None, dir2=None, angle=None):
        super().__init__(dir1, dir2, angle)



from .Solver import *

class LinearSolver(Solver):
    r"""
    A linear system solver

    This solver implies :math:`\{{\bf P}\} = [{\bf K}] \{{\bf u}\}`
    """

    def __init__(self):
        super().__init__()

    def solve(self, **kwargs):
        r"""
        Solves the system assuming the given load is the total load
        and the obtained displacement is the **total** displacement.

        .. note::

           This method will not verify wether or not the linear assumption is correct.
           The resulting forces may be out of equilibrium if the system experiences
           nonlinear behavior under the given load.

        """
        self.resetDisplacements()
        self.assemble()
        errorNorm = self.solveSingleStep()
        # recover residual force vector
        self.assemble(force_only=True)

        # time to add the information to the recorded data
        if self.record:
            self.recordThisStep()

        return errorNorm

    def solveSingleStep(self):

        # are we doing displacement control?
        if self.hasConstraint:

            # solve for displacement update: a single Newton step
            dQ = np.linalg.solve(self.Kt, np.stack([self.R, self.P]).T)

            if self.useArcLength:
                # arc-length control
                delU  = self.sysU - self.lastConverged['U']
                dload = self.loadfactor - self.lastConverged['lambda']
                # g = self.arclength2 - delU@delU - self.alpha * dload*dload * self.P@self.P

                denum = 2.*np.dot(dQ[:,1], delU) + 2.*self.alpha * dload * self.P@self.P

                dlam = ( self.g - 2.*np.dot(dQ[:,0], delU) ) / denum

            else:
                # displacement control
                # g = (self.targetU - np.dot(self.sysU, self.en))

                dlam = self.g - np.dot(dQ[:,0], self.en)
                dlam /= np.dot(dQ[:,1], self.en)

            self.loadfactor += dlam
            dU = dQ[:,0] + dlam * dQ[:,1]

        else:
            # solve for displacement update: a single Newton step
            dU = np.linalg.solve(self.Kt, self.R)

        self.U = dU

        # update nodal displacements
        for node in self.nodes:
            idxK = node.getIdx4DOFs()
            node._updateDisp(dU[idxK])


    def assemble(self, force_only=False):
        r"""
        inherited from :code:`Solver` class.
        """
        super(LinearSolver, self).assemble(force_only=force_only)

    def pushState(self, state):
        r"""
        Pushes :code:`state` to the solver.
        The solver will use that data to update it's internal state.

        .. list-table:: **state** is defined as a dictionary with the following contents:

            * - **nodes**
              - list of node pointers (required)
            * - **elements**
              - list of element pointers (required)
            * - **P0**
              - system vector of initial forces
            * - **Pref**
              - system vector of reference forces
            * - **u1**
              - system vector of current (converged) displacements
            * - **un**
              - system vector of previous (converged) displacements
            * - **lam1**
              - load level of current (converged) displacements
            * - **lamn**
              - load level of previous (converged) displacements

        :param state: state of the solver
        """
        super(LinearSolver, self).pushState(state)

        if 'lam1' in state:
            self.loadfactor = state['lam1']



import numpy as np

class Record():
    """
    container class for recording time history data
    """

    def __init__(self, key='', label=''):
        self.data  = []
        self.key   = key
        self.label = label

    def __str__(self):
        if len(self.data) > 3:
            s = "{}::{}::{}".format(self.label,self.key,self.data[:3])
        else:
            s = "{}::{}::{}".format(self.label,self.key,self.data)
        return s

    def __repr__(self):
        if len(self.data) > 3:
            s = "Record(label={}, key={}, data=[{},{},{},...])".format(self.label,self.key,*self.data[:3])
        else:
            s = "Record(label={}, key={}, data={})".format(self.label,self.key,self.data)
        return s

    def __len__(self):
        """
        :return: the length of the collected data array
        """
        return len(self.data)

    def isKey(self, key):
        return (key and key == self.key)

    def getData(self):
        """
        :return: tuple (label, data.asarray())
        """
        return (self.label, np.array(self.data))



"""
=====================================================
Diffusion elements
=====================================================

Elements in this group share the following features:
* gradients with respect to an undeformed configuration
* flux proportional to gradient of the potential.

"""

__all__ = (
    "Triangle",
    "Triangle6",
)

from .Triangle import *
from .Triangle6 import *




"""
Defines an Example class from which all provided examples will be derived.

"""


class Example():

    def __init__(self):
        pass

    def __str__(self):
        s  = "Example {}\n".format(self.__class__.__name__)
        #s += len(s)*"-" + '\n'
        doc_string = self.docString()
        if doc_string:
            s += doc_string + '\n'
        return s

    def run(self):
        self.problem()
        return

        try:
            self.problem()
        except:
            print("failure to run Example: {}".format(self.__class__.__name__))

    def docString(self):
        """
        Return problem specific documentation as a multi-line string.
        """
        return ""

    def problem(self, *args, **kwargs):
        """
        Build and run the example problem
        """

        msg = """
        Missing problem() method for Example {}
        """.format(self.__class__.__name__)
        print(msg)





from .Integration import Integration

class TriangleIntegration(Integration):
    r"""
    Provides integration points and weights for a triangular domain (2d)
    """

    def __init__(self, order=2):
        super(TriangleIntegration, self).__init__(order=order, dimension=1)

        xi, w = self.dunavant( p=order )
        self.xi = xi
        self.w  =  w



from .Constraint import Constraint

class SPconstraint(Constraint):
    """
    wrapper for a single point constraint
    """

    def __init__(self, node, dofs, a, u0, u1=0, lam0=0, **kwargs):
        """
        :param node: one :py:class:`Node` object
        :param dofs: list of dof codes for which to apply those constraints
        :param a: list or array of factors associated with each listed **dof**
        """
        params = {}
        params['u0']   = u0
        params['u1']   = u1
        params['lam0'] = lam0
        params['dofs'] = dofs
        params['a']    = a

        self.neqns  = 0

        super(SPconstraint, self).__init__([node], params, **kwargs)



__all__ = (
    'Material',
    'FiberMaterial',
    'PlaneStrain',
    'PlaneStress',
    'SectionMaterial',
    'ElasticSection',
    'DiffusionGeneral',
    'Thermal',
)

from .Material import *
from .FiberMaterial import *
from .PlaneStrain import *
from .PlaneStress import *
from .SectionMaterial import *
from .ElasticSection import *
from .DiffusionGeneral import *
from .Thermal import *



import numpy as np

from ..solver.Solver import Solver

class NewtonRaphsonSolver(Solver):
    r"""
    An iterative solver for load-controlled nonlinear analysis.
    """

    def __init__(self):
        super(NewtonRaphsonSolver, self).__init__()

    def solve(self, max_steps=10, verbose=False, **kwargs):
        r"""
        :param max_step: maximum number of iterations (int)
        :param verbose: set to :code:`True` for additional information
        """
        TOL = self.TOL

        if 'tol' in kwargs:
            TOL = kwargs['tol']
        if 'tolerance' in kwargs:
            TOL = kwargs['tolerance']

        for k in range(max_steps):

        # compute force vector and tangent stiffness
            self.assemble()

            # vals=(
            #     np.sum(self.P), np.linalg.norm(self.P),
            #     np.sum(self.R), np.linalg.norm(self.R)
            # )
            # print("assembly has sum(P)={:.3g}   norm(P)={:.3g}   sum(R)={:.3g}   norm(R)={:.3g}".format(*vals))

            # we are using , force_only=False and simply reuse self.Kt from that run (!)
            normR = self.checkResiduum(verbose, force_only=False)

            if normR < TOL:
                # we achieved convergence
                #
                # now broadcast that we can switch to this converged state
                self.on_converged()

                # time to add the information to the recorded data
                if self.record:
                    self.recordThisStep()

                break  # converged! break the iteration loop

            # Solve for equilibrium
            self.solveSingleStep()

        print('+')

        if normR > TOL:
            # we failed to converge
            #
            # back to safety: revert to the last converged step
            self.revert()

        return normR

    def solveSingleStep(self):
        r"""
        Helper function performing a single solution of the linearized system

        Called by **solve()**. (internal use only)
        """

        # are we doing displacement control?
        if self.hasConstraint:

            # solve for displacement update: a single Newton step
            dQ = np.linalg.solve(self.Kt, np.stack([self.R, self.P]).T)

            if self.useArcLength:
                # arc-length control
                dload = self.loadfactor - self.loadfactor_n
                denum = 2.*self.alpha * dload * self.P @ self.P
                numerator = self.g

                for node in self.nodes:
                    idx = node.getIdx4DOFs()
                    delU = node.getDeltaU()
                    denum     += 2.*np.dot(dQ[idx,1], delU)
                    numerator -= 2.*np.dot(dQ[idx,0], delU)

                dlam = numerator / denum

            else:
                # displacement control
                # g = (self.targetU - np.dot(self.sysU, self.en))
                idx = self.control_node.getIdx4DOFs(dofs=[self.control_dof])[0]

                dlam = self.g - dQ[idx,0]
                dlam /= dQ[idx,1]

            self.loadfactor += dlam
            dU = dQ[:,0] + dlam * dQ[:,1]

        else:
            # solve for displacement update: a single Newton step
            dU = np.linalg.solve(self.Kt, self.R)

        # update nodal displacements
        for node in self.nodes:
            idxK = node.getIdx4DOFs()
            node._updateDisp(dU[idxK])

    def assemble(self, force_only=False):
        r"""
        inherited from :code:`Solver` class.
        """
        super(NewtonRaphsonSolver, self).assemble(force_only=force_only)

    def getDisplacements(self):
        r"""
        .. warning::

            This method only works if all nodes have the same dof list (!)
        """
        U = self.sysU.copy()
        U.shape = (self.nNodes, self.ndof)
        return U

    def getForces(self):
        r"""
        .. warning::

            This method only works if all nodes have the same dof list (!)
        """
        P = self.P.copy() * self.loadfactor
        P.shape = (self.nNodes, self.ndof)
        return P

    def getResiduum(self):
        r"""
        ** NEEDS REDESIGN TO WORK WITH SMART NODES **
        """

        # R = self.R.copy()
        # R.shape = (self.nNodes, self.ndof)
        # return R

        return self.R.copy()

    def pushState(self, state):
        r"""
        Pushes :code:`state` to the solver.
        The solver will use that data to update it's internal state.

        .. list-table:: **state** is defined as a dictionary with the following contents:

            * - **nodes**
              - list of node pointers (required)
            * - **elements**
              - list of element pointers (required)
            * - **P0**
              - system vector of initial forces
            * - **Pref**
              - system vector of reference forces
            * - **u1**
              - system vector of current (converged) displacements
            * - **un**
              - system vector of previous (converged) displacements
            * - **lam1**
              - load level of current (converged) displacements
            * - **lamn**
              - load level of previous (converged) displacements

        :param state: state of the solver
        """
        super(NewtonRaphsonSolver, self).pushState( state)

        if 'lam1' in state:
            self.loadfactor = state['lam1']

    # arc-length control

    def initArcLength(self, load_increment=1., alpha=0.0, tolerance=1.0e-12):
        r"""
        Initializes parameters for the arc-length constraint.

        .. math::

           g({\bf u}, \lambda)) := \alpha ||\bar {\bf P}|| (\lambda-\lambda_n)^2 + ({\bf u} - {\bf u}_n)({\bf u} - {\bf u}_n) - \Delta s^2 = 0


        :param load_increment:   load increment used to calibrate the constraint
        :param alpha:            load contribution factor
        :param tolerance:        convergence tolerance
        """

        # store analysis parameter(s)
        self.alpha = alpha
        self.TOL   = tolerance

        # make sure we start at an equilibrium point
        self.solve()

        # use load control to solve for the new equilibrium state
        self.hasConstraint = False          # this forces load control
        self.loadfactor += load_increment   # add reference load level
        self.solve()                        # find equilibrium configuration for given load level

        # compute the arc-length for that step and store as target arc length
        g = self.alpha * load_increment**2 * self.P@self.P
        for node in self.nodes:
            g += node.getNormDeltaU2(previous_step=True)
        self.arclength2 = g                 # target arc-length

        # set solver parameters
        self.hasConstraint = True
        self.useArcLength  = True

    def stepArcLength(self, verbose=False, max_iter=10):
        r"""
        Progresses the model state by one arc-length.

        .. note::

            You need to initialize arc-length control by one call to
            :py:meth:`initArcLength` at least once to set all necessary parameters.

        :params max_iter: maximum number of iteration steps; handed on to the solver
        :return normR: the norm of the generalized residuum from the last iteration step
        """

        if not self.useArcLength:
            # this method makes no sense for load control
            return

        # set solver parameters
        self.hasConstraint = True
        self.useArcLength  = True

        # set suitable trial state
        for node in self.nodes:
            node.setTrialState()

        if verbose:
            print("Last converged state stored at lam={}".format(self.loadfactor_n))

        # solve for next point on the equilibrium path
        normR = self.solve(max_steps=max_iter, verbose=True)

        return (self.loadfactor, normR)



class GPdataType():
    """
    Gauss Point data type

    Holds:
    :param base: base vectors of local coordinate system
    :param dual_base: dual base vectors of local coordinate system
    :param state: state dictionary at current point. Holds strain, stress, etc.
    :param J: Jacobian (area or volume ratio) at current point
    :param B: B-matrix (kinematic matrix) at current point
    :param C: C-matrix (material tangent) at current point
    :param material: pointer to material object at current point
    :param Grad: gradient operator [ [ dN1/dx, dN1/dy], [ dN2/dx, dN2/dy], ... ]
    """

    def __init__(self):
        self.X         = None
        self.base      = None
        self.dual_base = None
        self.J         = 1.0
        self.B         = None
        self.state     = {}
        self.C         = 1.0
        self.material  = None
        self.Grad      = None

    def __str__(self):
        s  = f'GP @ X={self.base}\n'
        s += f'* base:\n\t{self.base}\n'
        s += f'* dual base:\n\t{self.dual_base}\n'
        s += f'* Jacobian (J):\n\t{self.J}\n'
        s += f'* state:\n\t{self.state}\n'
        s += f'* B:\n\t{self.B}\n'
        s += f'* C:\n\t{self.C}\n'
        s += f'* material:\n\t{repr(self.material)}'
        s += f'* Grad:\n\t{self.Grad}'
        return s


from .Faces import *

class Face2D(Faces):
    r"""
    Implementation representing one face of a 2D element

    The face may be defined by two nodes (liner interpolation)
    or three nodes (quadratic interpolation).
    The node order for quadratic elements is nodes at `[-1, 0, +1]`
    """


    def __init__(self, id, *nds, **kwargs):
        super(Face2D, self).__init__(id, *nds, **kwargs)

    def initialize(self):
        r"""
        This method computes area and normals for :py:class:`Face2D`.

        It is called by the constructor of the :py:class:`Faces` class.
        """
        if self.dim < 2:
            msg = "{} requires 2D or 3D points - {} dimensions given".format(self.__class__.__name__, self.dim)
            raise TypeError(msg)

        rot = np.array([[0.,1.],[-1.,0.]])

        if len(self.nodes) == 2:
            # this includes the weight factor for single point integration over [-1,+1]
            self.pos     = np.array( [ 0.5*(self.nodes[1].getPos() + self.nodes[0].getPos()) ] )
            self.tangent = np.array( [ self.nodes[1].getPos() - self.nodes[0].getPos() ] )
            self.area    = np.array( [ rot @ tvec for tvec in self.tangent ] )
        elif len(self.nodes) == 3:
            # this includes the weight factor for two-point integration over [-1,+1]
            x0 = 0.4553418012614795 * self.nodes[0].getPos() \
                 + 0.6666666666666667 * self.nodes[1].getPos() \
                 - 0.1220084679281462 * self.nodes[2].getPos()
            x1 = -0.1220084679281462 * self.nodes[0].getPos() \
                 - 0.6666666666666667 * self.nodes[1].getPos() \
                 + 0.4553418012614795 * self.nodes[2].getPos()
            t0 = -1.077350269189626 * self.nodes[0].getPos() \
                 + 1.154700538379252 * self.nodes[1].getPos() \
                 - 0.07735026918962576 * self.nodes[2].getPos()
            t1 = 0.07735026918962576 * self.nodes[0].getPos() \
                 - 1.154700538379252 * self.nodes[1].getPos() \
                 + 1.077350269189626 * self.nodes[2].getPos()
            self.pos     = np.array( [ x0, x1 ] )
            self.tangent = np.array( [ t0, t1 ] )
            self.area    = np.array( [ rot @ t0, rot @ t1 ] )
        else:
            msg = "{} requires 2 or 3 points - {} given".format(self.__class__.__name__, len(self.nodes))
            raise TypeError(msg)

    def computeNodalForces(self):
        r"""
        Implementation of nodal force calculation from surface loads.

        The surface is described  as

        .. math::

            {\boldsymbol\varphi}(s) = \sum_k \phi_k(s) {\bf X}_k

        where :math:`\phi_k(s)` is the k-th shape function on `[-1,1]`,
        and :math:`{\bf X}_k` is the coordinate of the k-th node.

        The nodal forces are obtained as

        .. math::

            {\bf P}_i = \int_{-1}^{+1} \phi_i(s)
            \left(
                p_n \, {\boldsymbol\varphi}_{,s} \times {\bf k}
                + p_s \, {\boldsymbol\varphi}_{,s}
            \right) ds

        """
        if len(self.nodes) == 2:
            pn = self.load[0]
            ps = self.load[1]
            dF = (pn * self.area + ps * self.tangent)
            forces = np.outer( np.array([ 0.50000, 0.50000 ]), dF )

        elif len(self.nodes) == 3:
            pn = self.load[0]
            ps = self.load[1]

            tractions = (pn * self.area + ps * self.tangent)  # surface traction * weights at gauss points
            forces = np.array([[ 0.4553418012614795, 0.6666666666666667, -0.1220084679281462 ],
                               [ -0.1220084679281462, 0.6666666666666667, 0.4553418012614795 ]]).T \
                     @ tractions

        else:
            msg = "{} requires 2 or 3 points - {} given".format(self.__class__.__name__, len(self.nodes))
            raise TypeError(msg)

        # for i, node in enumerate(self.nodes):
        #     node.setLoad(forces[i])

        return forces  # this may need a different approach to account for the global load factor

    def computeNodalFlux(self):
        r"""
        Implementation of nodal flux calculation from surface flux.

        Surface in-flux (out-flux) is considered positive (negative).

        The surface is described  as

        .. math::

            {\boldsymbol\varphi}(s) = \sum_k \phi_k(s) {\bf X}_k

        where :math:`\phi_k(s)` is the k-th shape function on `[-1,1]`,
        and :math:`{\bf X}_k` is the coordinate of the k-th node.

        The nodal fluxes are obtained as

        .. math::

            {\bf P}_i = \int_{-1}^{+1} \phi_i(s) \: q_n \: || {\boldsymbol\varphi}_{,s} || \: ds

        """
        if len(self.nodes) == 2:
            qn = self.flux
            dF = qn * np.linalg.norm(self.area)
            fluxes = np.array([ 0.50000, 0.50000 ]) * dF

        elif len(self.nodes) == 3:
            qn = self.flux

            dF = qn * np.linalg.norm(self.area)  # surface flux * weights at gauss points
            fluxes = np.array([ 0.1666666666666667, 0.6666666666666667, 0.1666666666666667 ]) * dF

        else:
            msg = "{} requires 2 or 3 points - {} given".format(self.__class__.__name__, len(self.nodes))
            raise TypeError(msg)

        # for i, node in enumerate(self.nodes):
        #     node.setLoad(forces[i])

        return fluxes  # this may need a different approach to account for the global load factor

    def isFace(self, X, N):
        r"""
        Implementation of the test function.

        :param X: position vector for a point
        :type X: np.array
        :param N: outward normal vector at **X**
        :type N: np.array
        :return: **True** if **X** and **N** match this face. **False** otherwise.
        """
        Lvec = self.nodes[-1].getPos() - self.nodes[0].getPos()
        L = np.linalg.norm(Lvec)

        if self.num_nodes == 2:
            xm = 0.5 * (self.nodes[-1].getPos() + self.nodes[0].getPos())
            Nvec = self.area[0]
        elif self.num_nodes == 3:
            xm = self.nodes[1].getPos()
            Nvec = self.area[0] + self.area[-1]
        else:
            msg = f"Face2D does not understand {self.num_nodes} nodes on one surface"
            raise ValueError(msg)

        # approximate projection
        area = np.linalg.norm(Nvec)
        Nvec /= area
        s = 2.0 * (X-xm) @ Lvec / (Lvec @ Lvec)
        t = (X-xm) @ Nvec

        dir = N @ Nvec / np.linalg.norm(N)
        if dir > 0.98:
            pass



        msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        raise NotImplementedError(msg)




from .Recorder import Recorder
from .NodeRecorder import *
from .ElementRecorder import *

class ModelRecorder(Recorder):
    """

    """

    def __init__(self,**kwargs):
        super(ModelRecorder, self).__init__(**kwargs)

        for var in self.sysvars:
            self.data[var] = Record(key=var, label=f"sys:{var}")

        self.nodes    = []
        self.elements = []

        if 'nodes' in kwargs:
            self.nodes = kwargs['nodes']
            for node in self.nodes:
                node.setRecorder(NodeRecorder(variables=self.nodevars, label=node.getID(), type='node'))

        if 'elements' in kwargs:
            self.elements = kwargs['elements']
            for elem in self.elements:
                elem.setRecorder(ElementRecorder(variables=self.elemvars, type='element'))




"""
Example: frame01

Buckling of a beam with pin-pin support

modeled using a 2D frame element

.. code::

    x============x=============x  <---
    ^                          o

    x ..... node
    === ... frame element
    <-- ... applied force
    ^ ..... pin support
    o ..... roller support

    degrees of freedom:
    0 ... horizontal displacement, u
    1 ... vertical displacement, v
    2 ... rotation, theta


.. list-table:: setting given parameters

    * - N  = 2
      - number of elements
    * - L  = 100.0
      - column length
    * - EA = 2000000.0
      - axial stiffness
    * - EI = 21000.0
      - flexural stiffness
    * - w  = 0.1
      - applied lateral load


"""
from ...examples.Example import *

from ...domain import *
from ...solver.NewtonRaphsonSolver import *
from ...elements.Frame2D import *
from ...materials.ElasticSection import *


class ExampleFrame01(Example):

    def docString(self):
        s = """
    Buckling of a beam with pin-pin support
    
    modeled using a 2D frame element
    
        x============x=============x  <---
        ^                          o
        
        x ..... node
        === ... frame element
        <-- ... applied force
        ^ ..... pin support
        o ..... roller support
        
    degrees of freedom:
        0 ... horizontal displacement, u
        1 ... vertical displacement, v
        2 ... rotation, theta
        
    parameters 
        
        N  = 2     # number of elements
        L  = 100.0
        EA = 2000000.0
        EI = 21000.0
        w  = 0.1
        
    Author: Peter Mackenzie-Helnwein 
    """
        return s

    def problem(self):

        #
        # ==== Initialization ====
        #

        # ========== setting mesh parameters ==============

        N = 8         # number of elements in the mesh
        L = 100.0     # column free length


        # ========== setting material parameters ==============

        params = dict(
            E = 20000.,    # Young's modulus
            A = 100.0,     # cross section area
            I = 10.0       # cross section moment of inertia
        )

        # ========== setting load parameters ==============

        w   = -0.1         # uniform lateral load on the column
        Pcr = np.pi**2 * params['E'] * params['I'] / L**2    # Euler buckling load

        # ========== setting analysis parameters ==============

        target_load_level = 0.99      # 99% of Euler load
        max_steps = 10                # solve max_steps points on the primary path


        w   *= 0.01
        Pcr *= 0.01
        target_load_level = 99.      # 99% of Euler load



        # define a list of target load levels
        load_levels = np.linspace(0, target_load_level, max_steps)

        #
        # ==== Build the system model ====
        #

        model = System()
        model.setSolver(NewtonRaphsonSolver())

        # create nodes

        nd0 = Node(0.0, 0.0)
        model += nd0

        ndi = nd0
        for i in range(N):
            # nodes
            ndj = Node( (i+1)*L/N, 0.0 )
            model += ndj

            # elements
            elem = Frame2D(ndi, ndj, ElasticSection(params))
            model += elem

            # ** apply the element portion of the reference load
            elem.setDistLoad(w)

            ndi = ndj    # jump to next element: make current end-node the next start-node

        # define support(s)

        nd0.fixDOF('ux', 'uy')    # horizontal support left end
        ndi.fixDOF('uy')          # vertical support right end

        # ==== complete the reference load ====

        # these are only nodal forces as part of the reference load
        # .. load only the upper node
        ndi.setLoad((-Pcr,), ('ux',))

        # show model information
        print(model)

        #
        # ==== perform the analysis ===
        #

        print("\n==== perform the analysis ===\n")

        # * apply the load in multiple smaller load steps

        # set up data recorder
        model.initRecorder()
        model.trackStability(True)

        # initialize the analysis:
        model.resetDisplacements()   # set U to all zeros
        model.setLoadFactor(0.0)     # define a known equilibrium solution

        model.startRecorder()

        detKt   = []
        lambdas = []

        # solve for all load_levels
        for loadfactor in load_levels:

            # define node X2 as the controled node; downward direction is prescribed:
            model.setLoadFactor(loadfactor)
            model.solve(verbose=True)

            # stability check
            lambdas.append(model.loadfactor)
            detKt.append(model.solver.checkStability())

            # report results
            print('+')
            #model.report()

            print("\n=== next load level ===\n")


        #
        # ==== create some nice plots ===
        #

        model.report()

        plt.plot(lambdas,detKt,'--*r')
        plt.grid(True)
        plt.xlabel('Load factor, $ \\lambda $')
        plt.ylabel("Stability index, $ {det}\: {\\bf K}_t $")
        plt.show()

        model.plot(factor=10.0, filename="frame1_deformed.png")

        model.beamValuePlot("F", filename="frame1_force.png")
        model.beamValuePlot("V", filename="frame1_shear.png")
        model.beamValuePlot("M", filename="frame1_moment.png")

        model.plotBucklingMode(factor=10., filename="frame1_buckling_mode0.png")



import numpy as np

from .Transformation import *

class CosseratTransformation(Transformation):
    r"""
    A Transformation including displacement dofs and rotational dofs
    """

    def __init__(self, dir1=None, dir2=None, axis=None):
        super().__init__(dir1, dir2, axis)

        # reset vectors for transformation
        self.known_vectors = []
        self.addVector(['ux','uy','uz'])
        self.addVector(['rx','ry','rz'])



import numpy as np
import scipy as sc
from copy import deepcopy

from .Node import Node
from ..elements import Element


class Transformation():
    r"""
    Nodal transformation class

    A transformation may be attached to Nodes or to Elements.

    - When attached to a Node, the transformation applies to
      all fixities and all loads defined at that node.
    - When attached to an element, the user may specify a node for which that transformation
      shall be used.  If no node is specified, the transformation will be applied at all
      attached nodes.

    """

    def __init__(self, dir1=None, dir2=None, axis=None):
        """

        :param dir1: raw first direction
        :param dir2: raw second direction
        :param axis: axial rotation vector
        """

        self.dir1  = dir1          # user input raw first direction
        self.dir2  = dir2          # user input raw second direction
        self.axis  = axis          # user input axial rotation vector

        self.known_vectors = []    # create a list of vectors that should be transformed
        self.clients = []          # holds pointers to client nodes/elements

        dir1_is_vector = (isinstance(dir1, np.ndarray) or isinstance(dir1, list))
        dir2_is_vector = (isinstance(dir2, np.ndarray) or isinstance(dir2, list))
        axis_is_vector = (isinstance(self.axis, np.ndarray) or isinstance(self.axis, list))

        have_dirs = dir1_is_vector and dir2_is_vector
        msg = ""
        model_dim = 0

        if have_dirs:
            if len(dir1) == len(dir2):
                model_dim = len(dir1)
            else:
                msg = "Cannot define transformation from vectors of mismatched length: len(dir1)={}  len(dir2)={}".format(len(dir1),len(dir2))
        elif axis_is_vector:
            model_dim = len(axis)
        else:
            msg = f"Cannot define transformation from axis={axis}. Need a vector but got {type(axis)}"

        if model_dim not in (2,3):
            msg = f"Transformation only defined for 2d and 3d but got model dimension {model_dim}"

        if msg:
            raise TypeError(msg)

        if have_dirs and model_dim == 2:
            self.init_from_dirs_2d(have_dirs)
        elif have_dirs and model_dim == 3:
            self.init_from_dirs_3d(have_dirs)

    def init_from_dirs_2d(self, have_dirs):

        if have_dirs:
            g1 = np.array(self.dir1)
            g1 /= np.linalg.norm(g1)
            g2 = np.array(self.dir2)
            g2 -= (g1@g2) * g1
            g2 /= np.linalg.norm(g2)
            self.T = np.vstack((g1,g2)).T
        elif np.isreal(self.axis):
            w3 = self.axis
            omega = np.array([[0., -w3],[w3, 0.]])
            self.T = sc.linalg.expm(omega)
        else:
            self.T = np.identity(2)

    def init_from_dirs_3d(self, have_dirs):

        axis_is_vector = (isinstance(self.axis, np.ndarray) or isinstance(self.axis, list))

        if have_dirs:
            g1 = np.array(self.dir1)
            g1 /= np.linalg.norm(g1)
            g2 = np.array(self.dir2)
            g2 -= (g1@g2) * g1
            g2 /= np.linalg.norm(g2)
            g3 = np.linalg.cross(g1,g2)
            self.T = np.vstack((g1,g2,g3)).T
        elif axis_is_vector:
            w1,w2,w3 = self.axis
            omega = np.array([[0., -w3, w2],[w3, 0., -w1],[ -w2, w1, 0.]])
            self.T = sc.linalg.expm(omega)
        else:
            self.T = np.identity(3)

    def __str__(self):
        s = "{}(".format(self.__class__.__name__)
        if np.any(self.dir1):
            s += f"dir1={self.dir1}, "
        if np.any(self.dir2):
            s += f"dir2={self.dir2}, "
        if np.any(self.axis):
            s += f"axis={self.axis}, "
        s += ")"
        return s

    def registerClient(self, client):
        """
        Register a Node or Element object with this transformation.

        :param client: pointer to the node/element to be registered
        """
        if isinstance(client, (Node, Element)):
            if client not in self.clients:
                self.clients.append(client)
                self.refreshMaps()
        else:
            msg = "Only Nodes and Elements can interact with a Transformation: {} given".format(type(client).__name__)
            raise TypeError(msg)

    def addVector(self, vecdef):
        """
        Register a vector to be transformed by this Transformation object.

        This method is most commonly used internally, though may be used to add user-defined
        vector dofs to any (nodal) transformation.

        :param vecdef: an array-like definitions of a vector.

        **Examples for vecdef**

        * 2d solids use :code:`['ux,'uy']`
        * 3d solids use :code:`['ux','uy','uz']`
        * 3d Frames use :code:`['ux','uy','uz']` and :code:`['rx','ry','rz']`
        * your own vector dofs may use :code:`['myx','myy','myz']` if 'myx' represents the x-component of a :code:`my`-vector. These dofs must be requested by the constructor of your element.
        """
        newvec = list(vecdef)  # ensure object is a list
        if newvec not in self.known_vectors:
            self.known_vectors.append({'target':newvec, 'map':None})
            self.refreshMaps()

    def refreshMaps(self):
        r"""
        Reinitialize dof-maps for vector transformations.

        .. note::

            This method shall be called every time a vector is registered for transformation **and**
            if an element is added to a node since a new element addition may change the list of
            locally available dofs!

        The map structure is defined as

        .. code::

            self.known_vectors = [
                # one entry per vector that needs transformation
                {
                    'dofs': [ dofs_defining_this_vector ],
                    'map': {
                                # one entry per client
                                Node|Element: [ idx_for_dof[0], idx_for_dof[1], ... ],
                                ...
                           }
                },
                ...
            ]

        The map for any requested vector can be obtained using

        .. code::

            node = ...     # pointer to node of interest
            elem = ...     # pointer to elem of interest

            for vec in self.known_vectors:

                vector_dofs = vec['dofs']  # list of named dofs forming this vector

                if node in vec['map']:
                    node_map = vec['map'][node]

                if elem in vec['map']:
                    elem_map = vec['map'][elem]

        """
        for vec in self.known_vectors:
            map = {}
            for client in self.clients:
                if isinstance(client, Node):
                    map[client] = client.getIdx4DOFs(dofs=vec['target'], local=True)
                elif isinstance(client, Element):
                    map[client] = client.getDofs()
            vec['map'] = map

    def getT(self):
        r"""
        returns the transformation matrix that maps components from a local 3d vector to the global system

        **Usage**

        .. code::

            T = transform.getT()

            # vectors
            Uglobal = T @ Ulocal          # local to global for a vector U
            Vlocal = T.T @ Vglobal        # global to local for a vector V
            Vlocal = Vglobal @ T          # alternative global to local for a vector V

            # matrices
            Mglobal = T @ Mlocal @ T.T    # local to global
            Mlocal  = T.T @ Mglobal @ T   # global to local

        """
        return self.T

    #
    # acting on 2d|3d vectors|matrices
    #

    def vectorToGlobal(self, U, dof_list=[]):
        r"""
        :param U: 2d or 3d vector
        """
        return U @ self.T

    def vectorToLocal(self, U, dof_list=[]):
        r"""
        :param U: 2d or 3d vector
        """
        return self.T @ U

    def matrixToGlobal(self, M, dof_list=[]):
        r"""
        :param M: [2x2] or [3x3] matrix
        """
        return self.T @ M @ self.T.T

    def matrixToLocal(self, M, dof_list=[]):
        r"""
        :param M: [2x2] or [3x3] matrix
        """
        return self.T.T @ M @ self.T

    #
    # acting on the entire Node vector
    #

    def v2g(self, U, caller=None):
        r"""
        :param U: [ndof] vector w/ ndof = number of dofs at Node||Element
        """
        Utransformed = deepcopy(U)

        maps = self._caller_check(caller)

        for map in maps:
            local_vec  = Utransformed[map]
            global_vec = self.T @ local_vec
            Utransformed[map] = global_vec

        return Utransformed

    def v2l(self, U, caller=None):
        r"""
        :param U: [ndof] vector w/ ndof = number of dofs at Node||Element
        """
        Utransformed = deepcopy(U)

        maps = self._caller_check(caller)

        for map in maps:
            local_vec  = Utransformed[map]
            global_vec = self.T.T @ local_vec
            Utransformed[map] = global_vec

        return Utransformed

    def m2g(self, M, caller=None):
        r"""
        :param M: [ndof x ndof] matrix w/ ndof = number of dofs at Node||Element
        """
        maps = self._caller_check(caller)

        return self.T @ M @ self.T.T

    def m2l(self, M, caller=None):
        r"""
        :param M: [ndof x ndof] matrix w/ ndof = number of dofs at Node||Element
        """
        Mtransformed = deepcopy(M)

        maps = self._caller_check(caller)

        for map in maps:
            map = np.array(map, dtype=int)
            local_vec  = Mtransformed[map,:]
            global_vec = self.T.T @ local_vec
            Mtransformed[map,:] = global_vec

        return Mtransformed

        # return self.T.T @ M @ self.T

    def _caller_check(self, caller):
        if not caller:
            msg = "caller must be provided when requesting a vector|matrix transformation from {}".format(self.__class__.__name__)
            raise TypeError(msg)

        maps = []

        for vec in self.known_vectors:
            if caller in vec['map']:
                maps.append(vec['map'][caller])

        return maps


if __name__ == "__main__":

    vecs = []
    vecs.append(np.array([1.,0.,0.]))
    vecs.append(np.array([0.,1.,0.]))
    vecs.append(np.array([0.,0.,1.]))

    matrices = []
    matrices.append(np.array([[1.,0.,0.],[0.,1.,0.],[0.,0.,1.]]))
    matrices.append(np.array([[10.,0.,0.],[0.,20.,0.],[0.,0.,30.]]))
    matrices.append(np.array([[3.,2.,1.],[2.,3.,2.],[1.,2.,3.]]))

    transforms = []
    transforms.append(Transformation(dir1=[1.,0.,0.], dir2=[0.,1.,0.]))   # identity transformation
    transforms.append(Transformation(axis=[0.,0.,np.pi/360]))             # rotate CCW by 1.0 degrees
    transforms.append(Transformation(axis=[0.,0.,np.pi/2]))             # rotate CCW by 90.0 degrees
    transforms.append(Transformation(dir1=[1.,1.,0.], dir2=[0.,1.,0.]))   # rotate CCW by 45 degrees
    transforms.append(Transformation(axis=[0.,0.,np.pi/4]))               # rotate CCW by 45 degrees (alt input)

    for k,T in enumerate(transforms):
        print(T)
        print("transformation map:\n", T.getT())
        for i,vec in enumerate(vecs):
            print(f"vec{i} : {vec}")
            print(f"vec{i} g->l {T.vectorToLocal(vec)}")
            print(f"vec{i} l->g {T.vectorToGlobal(vec)}")
        for i,mat in enumerate(matrices):
            print(f"mat{i} : \n{mat}")
            print(f"mat{i} g->l \n{T.matrixToLocal(mat)}")
            print(f"mat{i} l->g \n{T.matrixToGlobal(mat)}")
        print("---")


import numpy as np

from .CosseratTransformation import *

class FrameTransformation(CosseratTransformation):
    r"""
    A Transformation including displacement dofs and rotational dofs


    If vectors :math:`{\bf d}_1` and :math:`{\bf d}_2` are given, the local coordinate frame :math:`\{{\bf g}_x,{\bf g}_y\},{\bf g}_z\}` is defined as follows

    .. math::
        {\bf g}_x = \frac{1}{||{\bf d}_1||} \: {\bf d}_1

    .. math::
        \hat{\bf g}_y = {\bf d}_2 - ({\bf d}_2\cdot{\bf g}_x) {\bf g}_x

    .. math::
        {\bf g}_y = \frac{1}{||\hat{\bf g}_y||} \: \hat{\bf g}_y

    .. math::
        {\bf g}_z = {\bf g}_y \times {\bf g}_y

    If instead an axial vector, :math:`{\boldsymbol\omega}`, is given, the local coordinate frame is defined as

    .. math::
        {\bf g}_i = {\boldsymbol\Lambda} \: {\bf e}_i
        \qquad i=x,y,z

    where  :math:`\{{\bf e}_x,{\bf e}_y,{\bf e}_z\}` is the global cartesian coordinate frame
    and

    .. math::
        {\boldsymbol\Lambda} := \exp[{\boldsymbol\Omega}]

    is the rotation tensor, generated
    as the matrix exponent of the skew-symmetric tensor :math:`{\boldsymbol\Omega}`. The latter is uniquely
    defined by the following equivalence.

    .. math::
        {\boldsymbol\Omega} \cdot {\bf h} \equiv {\boldsymbol\omega} \times {\bf h}
        \qquad\text{for all}\quad {\bf h}\in\mathbb{R}^3

    :param dir1: in-plane vector, **dir1** == :math:`{\bf d}_1`, defining the local x-direction
    :param dir2: in-plane vector, **dir2** == :math:`{\bf d}_2`, used to define the local y-direction
    :param axis: axial vector, :math:`{\boldsymbol\omega}`, in degrees
    """

    def __init__(self, dir1=None, dir2=None, axis=None):
        super().__init__(dir1, dir2, axis)



from .Material import *

class PlaneStrain(Material):
    r"""
    class: representing a 2d Plane Stress Material

    """

    def __init__(self, params={'E':1.0, 't':1.0, 'nu':0.0, 'fy':1.0e30}):
        super().__init__(params = params)

        self._type = self.PLANESTRAIN | self.PLASTIC

        if 't' not in self.parameters:
            self.parameters['t']  = 1.0

        # initialize strain
        self.plastic_strain = np.zeros(3)
        self.setStrain({'xx':0.0, 'yy':0.0, 'xy':0.0})

    def getThickness(self):
        if 't' in  self.parameters:
            return self.parameters['t']
        else:
            return 1.0

    def setStrain(self, eps):
        r"""
        update state for a user provided axial strain value

        :param eps:  strain or strain tensor
        :return: n/a
        """
        self.strain = np.array([eps['xx'],eps['yy'],eps['xy']])

        # update stress state
        E  = self.parameters['E']
        t  = self.parameters['t']
        nu = self.parameters['nu']
        fy = self.parameters['fy']

        # default consistency parameter
        gamma = 0.0

        self.Et = E*t/(1. - nu*nu) * np.array([[1.,nu,0.],[nu,1.,0.],[0.,0.,(1.-nu)/2.]])

        Cinv = 1/(E*t) * np.array([[1.,-nu,0.],[-nu,1.,0.],[0.,0.,2.*(1.+nu)]])
        Phi  = np.array([[2.,-1.,0.],[-1.,2.,0.],[0.,0.,6.]])

        # elastic predictor
        self.sig = self.Et @ ( self.strain - self.plastic_strain )

        # check yield condition
        (sxx, syy, sxy) = self.sig
        f = self.sig @ Phi @ self.sig / 2. - (t*fy)**2

        gamma = 0.0
        Xi = self.Et

        # plastic corrector as needed
        if f >= 0.0:
            # loop
            r = Phi @ self.sig
            Xixr = Xi @ r
            gamma += f / (r @ Xixr)
            Xi = (Cinv + gamma * Phi).I
            self.sig = Xi @ ( self.strain - self.plastic_strain )
            self.Et  = Xi

            print("material entering plastic state")

        #print(4*'{:12.8e}  '.format(eps, f, self.plastic_strain, self.sig ))

    def getStrain(self):
        return self.sig / self.parameters['E'] + self.plastic_strain

    def updateState(self):
        # update state now that the global analysis has converged
        E  = self.parameters['E']
        t  = self.parameters['t']
        nu = self.parameters['nu']

        Cinv = 1 / (E * t) * np.array([[1., -nu, 0.], [-nu, 1., 0.], [0., 0., 2. * (1. + nu)]])
        self.plastic_strain = self.strain - Cinv @ self.sig


if __name__ == "__main__":

    import matplotlib.pyplot as plt

    # testing the Node class
    mat = PlaneStrain(params={'E':100.0, 'nu':0.0, 'fy':1.0})

    eps = 0.02 * np.sin( np.linspace(0, 2.*np.pi, 100) )

    sig = []
    Et = []
    for strain in eps:
        # update material strain state
        mat.setStrain(strain)
        #collect stress response
        sig.append(mat.getStress())
        #collect material tangent stiffness response
        Et.append(mat.getStiffness())

    fig, (ax1,ax2) = plt.subplots(2,1)

    ax1.plot(eps, sig, '-r', label='stress')
    ax1.grid(True)
    ax1.set_xlabel('strain $\\varepsilon$')
    ax1.set_ylabel('stress $\\sigma$')
    ax1.legend()

    ax2.plot(eps, Et, '-.b', label='tangent modulus')
    ax2.grid(True)
    ax2.set_xlabel('strain $\\varepsilon$')
    ax2.set_ylabel('tangent modulus $E_t$')
    ax2.legend()

    plt.show()
